#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utilities for outputs."""

import os
from typing import TYPE_CHECKING, Dict, Sequence
from uuid import uuid4

from zenml.client import Client
from zenml.constants import IN_MEMORY_ARTIFACT_URI_PREFIX
from zenml.logger import get_logger
from zenml.utils import string_utils

if TYPE_CHECKING:
    from zenml.config.step_configurations import Step
    from zenml.models import StepRunResponse
    from zenml.stack import Stack


logger = get_logger(__name__)


def generate_artifact_uri(
    artifact_store_path: str,
    step_run: "StepRunResponse",
    output_name: str,
) -> str:
    """Generates a URI for an output artifact.

    Args:
        artifact_store_path: The path of the artifact store in which the
            artifact will be stored.
        step_run: The step run that created the artifact.
        output_name: The name of the output in the step run for this artifact.

    Returns:
        The URI of the output artifact.
    """
    for banned_character in ["<", ">", ":", '"', "/", "\\", "|", "?", "*"]:
        output_name = output_name.replace(banned_character, "_")
    return os.path.join(
        artifact_store_path,
        step_run.name,
        output_name,
        str(step_run.id),
        str(uuid4())[:8],  # add random subfolder to avoid collisions
    )


def prepare_output_artifact_uris(
    step_run: "StepRunResponse",
    stack: "Stack",
    step: "Step",
    *,
    skip_artifact_materialization: bool = False,
) -> Dict[str, str]:
    """Prepares the output artifact URIs to run the current step.

    Args:
        step_run: The step run for which to prepare the artifact URIs.
        stack: The stack on which the pipeline is running.
        step: The step configuration.
        skip_artifact_materialization: Whether to skip artifact materialization.

    Raises:
        RuntimeError: If an artifact URI already exists.

    Returns:
        A dictionary mapping output names to artifact URIs.
    """
    artifact_store = stack.artifact_store
    output_artifact_uris: Dict[str, str] = {}

    for output_name in step.config.outputs.keys():
        substituted_output_name = string_utils.format_name_template(
            output_name, substitutions=step_run.config.substitutions
        )
        artifact_store_path = (
            IN_MEMORY_ARTIFACT_URI_PREFIX
            if skip_artifact_materialization
            else artifact_store.path
        )
        artifact_uri = generate_artifact_uri(
            artifact_store_path=artifact_store_path,
            step_run=step_run,
            output_name=substituted_output_name,
        )

        if not skip_artifact_materialization:
            if artifact_store.exists(artifact_uri):
                raise RuntimeError("Artifact already exists")
            artifact_store.makedirs(artifact_uri)
        output_artifact_uris[output_name] = artifact_uri
    return output_artifact_uris


def remove_artifact_dirs(artifact_uris: Sequence[str]) -> None:
    """Removes the artifact directories.

    Args:
        artifact_uris: URIs of the artifacts to remove the directories for.
    """
    # TODO: maybe keep non-empty dirs here, if the step saved intermediate
    # checkpoints?
    artifact_store = Client().active_stack.artifact_store
    for artifact_uri in artifact_uris:
        if artifact_store.isdir(artifact_uri):
            artifact_store.rmtree(artifact_uri)



"""Release [0.91.1].

Revision ID: 0.91.1
Revises: a5a17015b681
Create Date: 2025-11-11 01:00:37.460524

"""

# revision identifiers, used by Alembic.
revision = "0.91.1"
down_revision = "a5a17015b681"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Filesystem registry managing filesystem plugins."""

import re
from threading import Lock
from typing import TYPE_CHECKING, Dict, Type

from zenml.logger import get_logger

if TYPE_CHECKING:
    from zenml.io.filesystem import BaseFilesystem, PathType


logger = get_logger(__name__)


class FileIORegistry:
    """Registry of pluggable filesystem implementations."""

    def __init__(self) -> None:
        """Initialize the registry."""
        self._filesystems: Dict["PathType", Type["BaseFilesystem"]] = {}
        self._registration_lock = Lock()

    def register(self, filesystem_cls: Type["BaseFilesystem"]) -> None:
        """Register a filesystem implementation.

        Args:
            filesystem_cls: Subclass of `zenml.io.filesystem.Filesystem`.
        """
        with self._registration_lock:
            for scheme in filesystem_cls.SUPPORTED_SCHEMES:
                current_preferred = self._filesystems.get(scheme)
                if current_preferred is not None:
                    logger.debug(
                        "Overwriting previously registered filesystem for "
                        "scheme `%s`. Old class: %s, new class: %s",
                        scheme,
                        current_preferred.__name__,
                        filesystem_cls.__name__,
                    )

                self._filesystems[scheme] = filesystem_cls

    def get_filesystem_for_scheme(
        self, scheme: "PathType"
    ) -> Type["BaseFilesystem"]:
        """Get filesystem plugin for given scheme string.

        Args:
            scheme: The scheme to get the filesystem for.

        Returns:
            The filesystem plugin for the given scheme.

        Raises:
            ValueError: If no filesystem plugin is registered for the given
                scheme.
        """
        if isinstance(scheme, bytes):
            scheme = scheme.decode("utf-8")
        if scheme not in self._filesystems:
            raise ValueError(
                f"No file systems were found for the scheme: "
                f"{scheme}. Please make sure that you are using "
                f"the right path and the all the necessary "
                f"integrations are properly installed."
            )
        return self._filesystems[scheme]

    def get_filesystem_for_path(
        self, path: "PathType"
    ) -> Type["BaseFilesystem"]:
        """Get filesystem plugin for given path.

        Args:
            path: The path to get the filesystem for.

        Returns:
            The filesystem plugin for the given path.

        Raises:
            ValueError: If no filesystem plugin is registered for the given
                path.
        """
        # Assume local path by default, but extract filesystem prefix if available.
        if isinstance(path, str):
            path_bytes = path.encode("utf-8")
        elif isinstance(path, bytes):
            path_bytes = path
        else:
            raise ValueError("Invalid path type: %r." % path)
        result = re.match(b"^([a-z0-9]+://)", path_bytes)
        if result:
            scheme = result.group(1).decode("utf-8")
        else:
            scheme = ""
        return self.get_filesystem_for_scheme(scheme)


# Default global instance of the filesystem registry.
default_filesystem_registry = FileIORegistry()



"""remove workspace from globals [3b1776345020].

Revision ID: 3b1776345020
Revises: 0392807467dc
Create Date: 2025-02-13 15:57:38.255825

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3b1776345020"
down_revision = "0392807467dc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision.

    Raises:
        RuntimeError: If more than one workspace exists.
    """
    # If more than one workspace exists, we fail the migration because it
    # would mean merging together resources from different workspaces, which
    # can lead to naming conflicts.
    workspace_count_query = sa.text("SELECT COUNT(*) FROM workspace")
    connection = op.get_bind()
    workspace_count = connection.execute(
        workspace_count_query,
    ).scalar()
    assert isinstance(workspace_count, int)
    if workspace_count > 1:
        raise RuntimeError(
            "Your ZenML installation has more than just the default workspace "
            "configured. This migration removes the workspace scopes of all "
            "stacks, components, flavors, service connectors and secrets, "
            "which may lead to naming conflicts if multiple workspaces are "
            "present. Please delete all but the default workspace before "
            "running this migration."
        )

    with op.batch_alter_table("flavor", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_flavor_workspace_id_workspace", type_="foreignkey"
        )
        batch_op.drop_column("workspace_id")

    with op.batch_alter_table("secret", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_secret_workspace_id_workspace", type_="foreignkey"
        )
        batch_op.drop_column("workspace_id")

    with op.batch_alter_table("service_connector", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_service_connector_workspace_id_workspace", type_="foreignkey"
        )
        batch_op.drop_column("workspace_id")

    with op.batch_alter_table("stack", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_stack_workspace_id_workspace", type_="foreignkey"
        )
        batch_op.drop_column("workspace_id")

    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_stack_component_workspace_id_workspace", type_="foreignkey"
        )
        batch_op.drop_column("workspace_id")


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision.

    Raises:
        NotImplementedError: This migration is not reversible.
    """
    raise NotImplementedError("This migration is not reversible.")



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline compilation context."""

import contextvars
from typing import TYPE_CHECKING

from typing_extensions import Self

from zenml.utils import context_utils

if TYPE_CHECKING:
    from zenml.pipelines.pipeline_definition import Pipeline


class PipelineCompilationContext(context_utils.BaseContext):
    """Pipeline compilation context."""

    __context_var__ = contextvars.ContextVar("pipeline_compilation_context")

    def __init__(
        self,
        pipeline: "Pipeline",
    ) -> None:
        """Initialize the pipeline compilation context.

        Args:
            pipeline: The pipeline that is being compiled.
        """
        super().__init__()
        self._pipeline = pipeline

    @property
    def pipeline(self) -> "Pipeline":
        """The pipeline that is being compiled.

        Returns:
            The pipeline that is being compiled.
        """
        return self._pipeline

    def __enter__(self) -> Self:
        """Enter the pipeline compilation context.

        Raises:
            RuntimeError: If the pipeline compilation context has already been
                entered.

        Returns:
            The pipeline compilation context object.
        """
        if self._token is not None:
            raise RuntimeError(
                "Compiling a pipeline while another pipeline is being compiled "
                "is not allowed."
            )
        return super().__enter__()



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Initialization for the ZenML vLLM integration."""
from typing import List, Type
from zenml.integrations.integration import Integration
from zenml.stack import Flavor
from zenml.logger import get_logger
from zenml.integrations.constants import VLLM

VLLM_MODEL_DEPLOYER = "vllm"

logger = get_logger(__name__)


class VLLMIntegration(Integration):
    """Definition of vLLM integration for ZenML."""

    NAME = VLLM

    REQUIREMENTS = ["vllm>=0.6.0,<0.7.0", "openai>=1.0.0"]

    @classmethod
    def activate(cls) -> None:
        """Activates the integration."""
        from zenml.integrations.vllm import services

    @classmethod
    def flavors(cls) -> List[Type[Flavor]]:
        """Declare the stack component flavors for the vLLM integration.

        Returns:
            List of stack component flavors for this integration.
        """
        from zenml.integrations.vllm.flavors import VLLMModelDeployerFlavor

        return [VLLMModelDeployerFlavor]





#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Run:AI API client wrapper with typed responses."""

from typing import Any, List, NoReturn, Optional, cast

from pydantic import BaseModel, ConfigDict, Field
from runai.api_client import ApiClient
from runai.configuration import Configuration
from runai.models.training_creation_request import TrainingCreationRequest
from runai.runai_client import RunaiClient as RunapyClient

from zenml.logger import get_logger

logger = get_logger(__name__)


class RunAIClientError(Exception):
    """Base exception for Run:AI client errors."""


class RunAIAuthenticationError(RunAIClientError):
    """Raised when authentication with Run:AI fails."""


class RunAIConnectionError(RunAIClientError):
    """Raised when connection to Run:AI API fails."""


class RunAIProjectNotFoundError(RunAIClientError):
    """Raised when a Run:AI project cannot be found."""

    def __init__(self, project_name: str, available: List[str]) -> None:
        """Initialize the exception.

        Args:
            project_name: The project name that was not found.
            available: List of available project names.
        """
        self.project_name = project_name
        self.available = available
        super().__init__(
            f"Project '{project_name}' not found in Run:AI. "
            f"Available projects: {available}"
        )


class RunAIClusterNotFoundError(RunAIClientError):
    """Raised when a Run:AI cluster cannot be found."""

    def __init__(self, cluster_name: str, available: List[str]) -> None:
        """Initialize the exception.

        Args:
            cluster_name: The cluster name that was not found.
            available: List of available cluster names.
        """
        self.cluster_name = cluster_name
        self.available = available
        super().__init__(
            f"Cluster '{cluster_name}' not found in Run:AI. "
            f"Available clusters: {available}"
        )


class RunAIWorkloadNotFoundError(RunAIClientError):
    """Raised when a Run:AI workload cannot be found."""

    def __init__(self, workload_id: str) -> None:
        """Initialize the exception.

        Args:
            workload_id: The workload ID that was not found.
        """
        self.workload_id = workload_id
        super().__init__(f"Workload '{workload_id}' not found in Run:AI.")


class RunAIProject(BaseModel):
    """Typed representation of a Run:AI project."""

    id: str
    name: str
    cluster_id: Optional[str] = None


class RunAICluster(BaseModel):
    """Typed representation of a Run:AI cluster."""

    id: str
    name: str


class WorkloadSubmissionResult(BaseModel):
    """Result of submitting a workload to Run:AI."""

    workload_id: str
    workload_name: str


class RunAITrainingWorkload(BaseModel):
    """Typed representation of a Run:AI training workload."""

    model_config = ConfigDict(extra="allow", populate_by_name=True)

    id: Optional[str] = None
    workload_id: Optional[str] = Field(default=None, alias="workloadId")
    name: Optional[str] = None
    actual_phase: Optional[str] = Field(default=None, alias="actualPhase")
    status: Optional[str] = None


class RunAIClient:
    """Wrapper around the runapy SDK providing typed responses.

    This client encapsulates all Run:AI API interactions and provides
    typed dataclasses instead of raw dictionaries.
    """

    def __init__(
        self, client_id: str, client_secret: str, runai_base_url: str
    ) -> None:
        """Initialize the Run:AI client.

        Args:
            client_id: Run:AI client ID for authentication.
            client_secret: Run:AI client secret for authentication.
            runai_base_url: Run:AI control plane base URL.

        Raises:
            RunAIConnectionError: If connecting to Run:AI fails.
            RunAIAuthenticationError: If client configuration fails.
        """
        try:
            config = Configuration(
                client_id=client_id,
                client_secret=client_secret,
                runai_base_url=runai_base_url,
            )
            self._raw_client = self._create_raw_client(config)
        except Exception as exc:
            if self._is_connection_error(exc):
                raise RunAIConnectionError(
                    f"Failed to connect to Run:AI API ({type(exc).__name__}): "
                    f"{exc}. Verify your runai_base_url and network "
                    "connectivity."
                ) from exc
            raise RunAIAuthenticationError(
                f"Failed to initialize Run:AI client ({type(exc).__name__}): {exc}. "
                "Verify your client_id, client_secret, and runai_base_url are correct."
            ) from exc

    def _create_raw_client(self, config: Configuration) -> RunapyClient:
        """Create the underlying runapy client.

        The runai SDK does not ship type hints, so keep the untyped call
        isolated to this helper.

        Args:
            config: The Configuration object to create the client from.

        Returns:
            The raw runapy client.
        """
        return RunapyClient(ApiClient(config))  # type: ignore[no-untyped-call]

    @staticmethod
    def _get_status_code(exc: Exception) -> Optional[int]:
        """Extract an HTTP status code from an untyped Run:AI SDK exception.

        Args:
            exc: The exception to extract the status code from.

        Returns:
            The HTTP status code, or None if not found.
        """
        for attr in ("status", "status_code", "code"):
            value = getattr(exc, attr, None)
            if isinstance(value, int):
                return value
        return None

    @staticmethod
    def _iter_exception_chain(exc: Exception) -> List[Exception]:
        """Iterate over an exception and its nested causes/contexts.

        Args:
            exc: The exception to inspect.

        Returns:
            A list containing the exception chain from outermost to innermost.
        """
        chain: List[Exception] = []
        visited: set[int] = set()
        current: Optional[Exception] = exc

        while current is not None and id(current) not in visited:
            chain.append(current)
            visited.add(id(current))
            next_exc = current.__cause__ or current.__context__
            current = next_exc if isinstance(next_exc, Exception) else None

        return chain

    @classmethod
    def _is_connection_error(cls, exc: Exception) -> bool:
        """Check whether an SDK exception represents a connectivity failure.

        Args:
            exc: The exception to classify.

        Returns:
            True if the exception indicates a connection/network issue.
        """
        connection_tokens = (
            "connection refused",
            "failed to establish a new connection",
            "max retries exceeded",
            "timed out",
            "temporary failure in name resolution",
            "name or service not known",
            "nodename nor servname provided",
            "no route to host",
            "connection reset by peer",
        )

        for current_exc in cls._iter_exception_chain(exc):
            if isinstance(current_exc, (ConnectionError, TimeoutError)):
                return True

            exception_name = type(current_exc).__name__.lower()
            if "connection" in exception_name or "timeout" in exception_name:
                return True

            message = str(current_exc).lower()
            if any(token in message for token in connection_tokens):
                return True

        return False

    @classmethod
    def _raise_api_error(cls, exc: Exception, message: str) -> NoReturn:
        """Raise a typed API error based on the underlying failure.

        Args:
            exc: The original exception.
            message: The error message to surface.

        Raises:
            RunAIConnectionError: If the root cause is connectivity related.
            RunAIClientError: For all other failures.
        """
        if cls._is_connection_error(exc):
            raise RunAIConnectionError(message) from exc
        raise RunAIClientError(message) from exc

    @property
    def raw_client(self) -> RunapyClient:
        """Access the underlying runapy client for advanced operations.

        Returns:
            The raw runapy client.
        """
        return self._raw_client

    def get_projects(self, search: Optional[str] = None) -> List[RunAIProject]:
        """Get Run:AI projects, optionally filtered by name.

        Args:
            search: Optional search string to filter projects.

        Returns:
            List of RunAIProject objects.
        """
        try:
            response = self._raw_client.organizations.projects.get_projects(
                search=search
            )
            projects_data = (
                response.data.get("projects", []) if response.data else []
            )

            return [
                RunAIProject(
                    id=str(p.get("id")),
                    name=p.get("name", ""),
                    cluster_id=p.get("clusterId"),
                )
                for p in projects_data
            ]
        except Exception as exc:
            self._raise_api_error(
                exc,
                f"Failed to fetch Run:AI projects "
                f"({type(exc).__name__}): {exc}",
            )

    def get_project_by_name(self, name: str) -> RunAIProject:
        """Get a Run:AI project by exact name match.

        Args:
            name: The project name to find.

        Returns:
            The matching RunAIProject.

        Raises:
            RunAIProjectNotFoundError: If no project matches the name.
        """
        projects = self.get_projects(search=name)

        for project in projects:
            if project.name == name:
                return project

        available = [p.name for p in projects]
        raise RunAIProjectNotFoundError(name, available)

    def get_clusters(self) -> List[RunAICluster]:
        """Get all Run:AI clusters.

        Returns:
            List of RunAICluster objects.
        """
        try:
            response = self._raw_client.organizations.clusters.get_clusters()
            clusters_data = response.data if response.data else []

            return [
                RunAICluster(
                    id=c.get("uuid", c.get("id", "")),
                    name=c.get("name", ""),
                )
                for c in clusters_data
            ]
        except Exception as exc:
            self._raise_api_error(
                exc,
                f"Failed to fetch Run:AI clusters "
                f"({type(exc).__name__}): {exc}",
            )

    def get_cluster_by_id(self, cluster_id: str) -> RunAICluster:
        """Get a Run:AI cluster by ID.

        Args:
            cluster_id: The cluster ID to find.

        Returns:
            The matching RunAICluster.

        Raises:
            RunAIClusterNotFoundError: If no cluster matches the ID.
        """
        clusters = self.get_clusters()
        for cluster in clusters:
            if cluster.id == cluster_id:
                return cluster
        available = [c.id for c in clusters]
        raise RunAIClusterNotFoundError(cluster_id, available)

    def get_cluster_by_name(self, name: str) -> RunAICluster:
        """Get a Run:AI cluster by exact name match.

        Args:
            name: The cluster name to find.

        Returns:
            The matching RunAICluster.

        Raises:
            RunAIClusterNotFoundError: If no cluster matches the name.
        """
        clusters = self.get_clusters()

        for cluster in clusters:
            if cluster.name == name:
                return cluster

        available = [c.name for c in clusters]
        raise RunAIClusterNotFoundError(name, available)

    def get_first_cluster(self) -> RunAICluster:
        """Get the first available Run:AI cluster.

        Returns:
            The first RunAICluster.

        Raises:
            RunAIClientError: If no clusters are available.
        """
        clusters = self.get_clusters()
        if not clusters:
            raise RunAIClientError("No Run:AI clusters available")
        return clusters[0]

    def create_training_workload(
        self, request: TrainingCreationRequest
    ) -> WorkloadSubmissionResult:
        """Submit a training workload to Run:AI.

        Args:
            request: TrainingCreationRequest from runai.models.

        Returns:
            WorkloadSubmissionResult with the workload ID.
        """
        try:
            response = self._raw_client.workloads.trainings.create_training1(
                training_creation_request=request
            )
            workload_id = self._extract_workload_id(response)
            return WorkloadSubmissionResult(
                workload_id=workload_id or request.name,
                workload_name=request.name,
            )
        except Exception as exc:
            self._raise_api_error(
                exc,
                f"Failed to submit Run:AI workload "
                f"({type(exc).__name__}): {exc}",
            )

    def _extract_workload_id(self, response: Any) -> Optional[str]:
        """Extract workload ID from API response.

        Args:
            response: The API response object.

        Returns:
            The workload ID or None.
        """
        if not response.data:
            return None

        if isinstance(response.data, dict):
            return (
                response.data.get("workloadId")
                or response.data.get("id")
                or response.data.get("workload_id")
            )
        return getattr(response.data, "workloadId", None) or getattr(
            response.data, "id", None
        )

    def get_training_workload_status(self, workload_id: str) -> Optional[str]:
        """Get the status of a training workload.

        Args:
            workload_id: The workload ID to query.

        Returns:
            The workload status string, or None if the response is missing a
            status field.

        Raises:
            RunAIWorkloadNotFoundError: If the workload was not found (404).
            RunAIClientError: If the API call fails for other reasons or the
                response is malformed.
        """
        try:
            response = self._raw_client.workloads.trainings.get_training(
                workload_id
            )
            if not response.data:
                raise RunAIClientError(
                    f"Empty response when querying workload {workload_id}. "
                    "The API returned no data."
                )
            if not isinstance(response.data, dict):
                raise RunAIClientError(
                    f"Unexpected response format for workload {workload_id}. "
                    f"Expected dict, got {type(response.data).__name__}."
                )
            status = response.data.get("actualPhase") or response.data.get(
                "status"
            )
            if status is None:
                logger.warning(
                    f"Workload {workload_id} response has no status field. "
                    "Available keys: %s",
                    list(response.data.keys()),
                )
            return cast(Optional[str], status)
        except RunAIClientError:
            raise
        except Exception as exc:
            status_code = self._get_status_code(exc)
            if status_code == 404:
                raise RunAIWorkloadNotFoundError(workload_id) from exc
            error_msg = str(exc).lower()
            if "not found" in error_msg or "404" in error_msg:
                raise RunAIWorkloadNotFoundError(workload_id) from exc
            self._raise_api_error(
                exc,
                f"Failed to query workload status "
                f"({type(exc).__name__}): {exc}",
            )

    def get_training_workload(self, workload_id: str) -> RunAITrainingWorkload:
        """Get full training workload details.

        Args:
            workload_id: The workload ID to query.

        Returns:
            The workload details as a typed model.

        Raises:
            RunAIClientError: If the query fails or response is invalid.
        """
        try:
            response = self._raw_client.workloads.trainings.get_training(
                workload_id
            )
            if not response.data:
                raise RunAIClientError(
                    f"Empty response when querying workload {workload_id}. "
                    "The workload may not exist or the API returned no data."
                )
            if not isinstance(response.data, dict):
                raise RunAIClientError(
                    f"Unexpected response format for workload {workload_id}. "
                    f"Expected dict, got {type(response.data).__name__}."
                )
            return RunAITrainingWorkload.model_validate(response.data)
        except RunAIClientError:
            raise
        except Exception as exc:
            self._raise_api_error(
                exc,
                "Failed to query Run:AI workload "
                f"{workload_id} ({type(exc).__name__}): {exc}",
            )

    def delete_training_workload(self, workload_id: str) -> None:
        """Delete a training workload.

        Args:
            workload_id: The workload ID to delete.
        """
        try:
            self._raw_client.workloads.trainings.delete_training(workload_id)
        except Exception as exc:
            self._raise_api_error(
                exc,
                "Failed to delete Run:AI workload "
                f"{workload_id} ({type(exc).__name__}): {exc}",
            )

    def suspend_training_workload(self, workload_id: str) -> None:
        """Suspend a training workload.

        Args:
            workload_id: The workload ID to suspend.
        """
        try:
            self._raw_client.workloads.trainings.suspend_training(workload_id)
        except Exception as exc:
            self._raise_api_error(
                exc,
                "Failed to suspend Run:AI workload "
                f"{workload_id} ({type(exc).__name__}): {exc}",
            )



"""Run wait conditions [24771f985eac].

Revision ID: 24771f985eac
Revises: 0.94.0
Create Date: 2026-03-13 15:16:21.636758

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "24771f985eac"
down_revision = "0.94.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "run_wait_condition",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("run_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("project_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=True),
        sa.Column(
            "resolved_by_user_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
        ),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column(
            "status", sqlmodel.sql.sqltypes.AutoString(), nullable=False
        ),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column(
            "question", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
        sa.Column("data_schema_json", sa.TEXT(), nullable=True),
        sa.Column(
            "resolution", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
        sa.Column("result_json", sa.TEXT(), nullable=True),
        sa.Column("last_polled_at", sa.DateTime(), nullable=True),
        sa.Column(
            "poller_instance_id",
            sqlmodel.sql.sqltypes.AutoString(),
            nullable=True,
        ),
        sa.Column("poller_lease_expires_at", sa.DateTime(), nullable=True),
        sa.Column("resolved_at", sa.DateTime(), nullable=True),
        sa.CheckConstraint(
            "(status != 'resolved') OR (resolution IS NOT NULL)",
            name="ck_run_wait_condition_resolved_requires_resolution",
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_run_wait_condition_project_id_project",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_run_wait_condition_user_id_user",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["resolved_by_user_id"],
            ["user.id"],
            name="fk_run_wait_condition_resolved_by_user_id_user",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["run_id"],
            ["pipeline_run.id"],
            name="fk_run_wait_condition_run_id_pipeline_run",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "run_id", "name", name="unique_wait_condition_name_per_run"
        ),
    )
    with op.batch_alter_table("run_wait_condition", schema=None) as batch_op:
        batch_op.create_index(
            "ix_run_wait_condition_project_id_created",
            ["project_id", "created"],
            unique=False,
        )
        batch_op.create_index(
            "ix_run_wait_condition_run_id_status",
            ["run_id", "status"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("run_wait_condition", schema=None) as batch_op:
        batch_op.drop_index("ix_run_wait_condition_run_id_status")
        batch_op.drop_index("ix_run_wait_condition_project_id_created")

    op.drop_table("run_wait_condition")
    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Functionality to generate stack component CLI commands."""

import time
from importlib import import_module
from typing import Any, Callable, List, Optional, Tuple, cast
from uuid import UUID

import click
from rich.markdown import Markdown

from zenml.cli import utils as cli_utils
from zenml.cli.annotator import register_annotator_subcommands
from zenml.cli.cli import TagGroup, cli
from zenml.cli.feature import register_feature_store_subcommands
from zenml.cli.model_registry import register_model_registry_subcommands
from zenml.cli.served_model import register_model_deployer_subcommands
from zenml.cli.utils import (
    OutputFormat,
    _component_display_name,
    get_default_output_format,
    is_sorted_or_filtered,
    list_options,
    print_model_url,
)
from zenml.client import Client
from zenml.console import console
from zenml.enums import CliCategories, StackComponentType
from zenml.exceptions import AuthorizationException, IllegalOperationError
from zenml.io import fileio
from zenml.models import (
    ComponentFilter,
    ServiceConnectorResourcesModel,
)
from zenml.utils import source_utils
from zenml.utils.dashboard_utils import get_component_url


def generate_stack_component_get_command(
    component_type: StackComponentType,
) -> Callable[[], None]:
    """Generates a `get` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """

    def get_stack_component_command() -> None:
        """Prints the name of the active component."""
        client = Client()
        display_name = _component_display_name(component_type)

        with console.status(f"Getting the active `{display_name}`...\n"):
            active_stack = client.active_stack_model
            components = active_stack.components.get(component_type, None)

            if components:
                cli_utils.declare(
                    f"Active {display_name}: '{components[0].name}'"
                )
                print_model_url(get_component_url(components[0]))
            else:
                cli_utils.warning(
                    f"No {display_name} set for active stack "
                    f"('{active_stack.name}')."
                )

    return get_stack_component_command


def generate_stack_component_describe_command(
    component_type: StackComponentType,
) -> Callable[..., None]:
    """Generates a `describe` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """

    @click.option(
        "--output",
        "-o",
        "output_format",
        type=click.Choice(["table", "json", "yaml", "csv", "tsv"]),
        default=None,
        help="Output format for the describe output. Uses ZENML_DEFAULT_OUTPUT env var if set, otherwise defaults to 'table'.",
    )
    @click.argument(
        "name_id_or_prefix",
        type=str,
        required=False,
    )
    def describe_stack_component_command(
        name_id_or_prefix: Optional[str] = None,
        output_format: Optional[OutputFormat] = None,
    ) -> None:
        """Prints details about the active/specified component.

        Args:
            name_id_or_prefix: Name or id of the component to describe.
            output_format: Output format (table, json, yaml, csv, tsv).
        """
        if output_format is None:
            output_format = get_default_output_format()

        client = Client()
        try:
            component_ = client.get_stack_component(
                name_id_or_prefix=name_id_or_prefix,
                component_type=component_type,
            )
        except KeyError as err:
            cli_utils.exception(err)

        with console.status(f"Describing component '{component_.name}'..."):
            active_component_id = None
            active_components = client.active_stack_model.components.get(
                component_type, None
            )
            if active_components:
                active_component_id = active_components[0].id

            if component_.connector:
                connector_requirements = (
                    component_.flavor.connector_requirements
                )
            else:
                connector_requirements = None

            url = get_component_url(component_)
            cli_utils.print_stack_component_configuration(
                component=component_,
                active_status=component_.id == active_component_id,
                connector_requirements=connector_requirements,
                output_format=output_format,
                dashboard_url=url if output_format != "table" else None,
            )

            if output_format == "table":
                print_model_url(url)

    return describe_stack_component_command


def generate_stack_component_list_command(
    component_type: StackComponentType,
) -> Callable[..., None]:
    """Generates a `list` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """

    @list_options(
        ComponentFilter,
        default_columns=["active", "id", "name", "flavor", "owner"],
    )
    @click.pass_context
    def list_stack_components_command(
        ctx: click.Context,
        /,
        columns: str,
        output_format: OutputFormat,
        **kwargs: Any,
    ) -> None:
        """Prints a table of stack components.

        Args:
            ctx: The click context object
            columns: Columns to display in output.
            output_format: Format for output (table/json/yaml/csv/tsv).
            kwargs: Keyword arguments to filter the components.
        """
        client = Client()
        with console.status(f"Listing {component_type.plural}..."):
            kwargs["type"] = component_type
            components = client.list_stack_components(**kwargs)

            show_active = not is_sorted_or_filtered(ctx)
            if show_active and components.items:
                active_stack = client.active_stack_model
                active_component = None
                if component_type in active_stack.components.keys():
                    active_components = active_stack.components[component_type]
                    active_component = (
                        active_components[0] if active_components else None
                    )

                if active_component is not None:
                    active_component_id = active_component.id
                    if active_component_id not in {
                        c.id for c in components.items
                    }:
                        components.items.insert(0, active_component)
                    components.items.sort(
                        key=lambda c: c.id != active_component_id
                    )
                else:
                    active_component_id = None
            else:
                active_component_id = None

            cli_utils.print_page(
                components,
                columns,
                output_format,
                empty_message="No components found for the given filters.",
                row_generator=cli_utils.generate_component_row,
                active_id=active_component_id,
            )

    return list_stack_components_command


def generate_stack_component_register_command(
    component_type: StackComponentType,
) -> Callable[[str, str, List[str]], None]:
    """Generates a `register` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "name",
        type=str,
    )
    @click.option(
        "--flavor",
        "-f",
        "flavor",
        help=f"The flavor of the {display_name} to register.",
        required=True,
        type=str,
    )
    @click.option(
        "--label",
        "-l",
        "labels",
        help="Labels to be associated with the component, in the form "
        "-l key1=value1 -l key2=value2.",
        multiple=True,
    )
    @click.option(
        "--connector",
        "-c",
        "connector",
        help="Use this flag to connect this stack component to a service connector.",
        type=str,
    )
    @click.option(
        "--resource-id",
        "-r",
        "resource_id",
        help="The resource ID to use with the connector. Only required for "
        "multi-instance connectors that are not already configured with a "
        "particular resource ID.",
        required=False,
        type=str,
    )
    @click.option(
        "--secret",
        "secrets",
        help="Secrets to attach to the component.",
        type=str,
        required=False,
        multiple=True,
    )
    @click.option(
        "--env",
        "environment_variables",
        help="Environment variables to set when running on this component. "
        "Must be of the format 'KEY=VALUE'.",
        type=str,
        required=False,
        multiple=True,
    )
    @click.argument("args", nargs=-1, type=click.UNPROCESSED)
    def register_stack_component_command(
        name: str,
        flavor: str,
        args: List[str],
        labels: Optional[List[str]] = None,
        connector: Optional[str] = None,
        resource_id: Optional[str] = None,
        secrets: List[str] = [],
        environment_variables: List[str] = [],
    ) -> None:
        """Registers a stack component.

        Args:
            name: Name of the component to register.
            flavor: Flavor of the component to register.
            args: Additional arguments to pass to the component.
            labels: Labels to be associated with the component.
            connector: Name of the service connector to connect the component to.
            resource_id: The resource ID to use with the connector.
            secrets: Secrets to attach to the component.
            environment_variables: Environment variables to set when running
                on this component.
        """
        client = Client()

        # Parse the given args
        # name is guaranteed to be set by parse_name_and_extra_arguments
        name, parsed_args = cli_utils.parse_name_and_extra_arguments(  # type: ignore[assignment]
            list(args) + [name], expand_args=True
        )

        parsed_labels = cli_utils.get_parsed_labels(labels)

        if connector:
            try:
                client.get_service_connector(connector)
            except KeyError as err:
                cli_utils.error(
                    f"Could not find a connector '{connector}': {str(err)}"
                )

        environment = {}
        for environment_variable in environment_variables:
            key, value = environment_variable.split("=", 1)
            environment[key] = value

        with console.status(f"Registering {display_name} '{name}'...\n"):
            # Create a new stack component model
            component = client.create_stack_component(
                name=name,
                flavor=flavor,
                component_type=component_type,
                configuration=parsed_args,
                labels=parsed_labels,
                secrets=secrets,
                environment=environment,
            )

            cli_utils.declare(
                f"Successfully registered {component.type} `{component.name}`."
            )
            print_model_url(get_component_url(component))

        if connector:
            connect_stack_component_with_service_connector(
                component_type=component_type,
                name_id_or_prefix=name,
                connector=connector,
                interactive=False,
                no_verify=False,
                resource_id=resource_id,
            )

    return register_stack_component_command


def generate_stack_component_update_command(
    component_type: StackComponentType,
) -> Callable[[str, List[str]], None]:
    """Generates an `update` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "name_id_or_prefix",
        type=str,
        required=False,
    )
    @click.option(
        "--label",
        "-l",
        "labels",
        help="Labels to be associated with the component, in the form "
        "-l key1=value1 -l key2=value2.",
        multiple=True,
    )
    @click.option(
        "--secret",
        "secrets",
        help="Secrets to attach to the component.",
        type=str,
        required=False,
        multiple=True,
    )
    @click.option(
        "--remove-secret",
        "remove_secrets",
        help="Secrets to remove from the component.",
        type=str,
        required=False,
        multiple=True,
    )
    @click.option(
        "--env",
        "environment_variables",
        help="Environment variables to set when running on this component. "
        "Must be of the format 'KEY=VALUE'. To remove an environment variable "
        "from the component, use an empty value, e.g. 'KEY='",
        type=str,
        required=False,
        multiple=True,
    )
    @click.argument("args", nargs=-1, type=click.UNPROCESSED)
    def update_stack_component_command(
        name_id_or_prefix: Optional[str],
        args: List[str],
        labels: Optional[List[str]] = None,
        secrets: List[str] = [],
        remove_secrets: List[str] = [],
        environment_variables: List[str] = [],
    ) -> None:
        """Updates a stack component.

        Args:
            name_id_or_prefix: The name or id of the stack component to update.
            args: Additional arguments to pass to the update command.
            labels: Labels to be associated with the component.
            secrets: Secrets to attach to the component.
            remove_secrets: Secrets to remove from the component.
            environment_variables: Environment variables to set when running
                on this component.
        """
        client = Client()

        # Parse the given args
        args = list(args)
        if name_id_or_prefix:
            args.append(name_id_or_prefix)

        name_or_id, parsed_args = cli_utils.parse_name_and_extra_arguments(
            args,
            expand_args=True,
            name_mandatory=False,
        )

        parsed_labels = cli_utils.get_parsed_labels(labels)

        environment = {}
        for environment_variable in environment_variables:
            key, value = environment_variable.split("=", 1)
            # Fallback to None if the value is empty so the existing environment
            # variable is removed
            environment[key] = value or None

        with console.status(f"Updating {display_name}...\n"):
            try:
                updated_component = client.update_stack_component(
                    name_id_or_prefix=name_or_id,
                    component_type=component_type,
                    configuration=parsed_args,
                    labels=parsed_labels,
                    add_secrets=secrets,
                    remove_secrets=remove_secrets,
                    environment=environment,
                )
            except KeyError as err:
                cli_utils.exception(err)

            cli_utils.declare(
                f"Successfully updated {display_name} "
                f"`{updated_component.name}`."
            )
            print_model_url(get_component_url(updated_component))

    return update_stack_component_command


def generate_stack_component_remove_attribute_command(
    component_type: StackComponentType,
) -> Callable[[str, List[str]], None]:
    """Generates `remove_attribute` command for a specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "name_id_or_prefix",
        type=str,
        required=True,
    )
    @click.option(
        "--label",
        "-l",
        "labels",
        help="Labels to be removed from the component.",
        multiple=True,
    )
    @click.argument("args", nargs=-1, type=click.UNPROCESSED)
    def remove_attribute_stack_component_command(
        name_id_or_prefix: str,
        args: List[str],
        labels: Optional[List[str]] = None,
    ) -> None:
        """Removes one or more attributes from a stack component.

        Args:
            name_id_or_prefix: The name of the stack component to remove the
                attribute from.
            args: Additional arguments to pass to the remove_attribute command.
            labels: Labels to be removed from the component.
        """
        client = Client()

        with console.status(
            f"Updating {display_name} '{name_id_or_prefix}'...\n"
        ):
            try:
                updated_component = client.update_stack_component(
                    name_id_or_prefix=name_id_or_prefix,
                    component_type=component_type,
                    configuration={k: None for k in args},
                    labels={k: None for k in labels} if labels else None,
                )
            except (KeyError, IllegalOperationError) as err:
                cli_utils.exception(err)

            cli_utils.declare(
                f"Successfully updated {display_name} `{name_id_or_prefix}`."
            )
            print_model_url(get_component_url(updated_component))

    return remove_attribute_stack_component_command


def generate_stack_component_rename_command(
    component_type: StackComponentType,
) -> Callable[[str, str], None]:
    """Generates a `rename` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "name_id_or_prefix",
        type=str,
        required=True,
    )
    @click.argument(
        "new_name",
        type=str,
        required=True,
    )
    def rename_stack_component_command(
        name_id_or_prefix: str, new_name: str
    ) -> None:
        """Rename a stack component.

        Args:
            name_id_or_prefix: The name of the stack component to rename.
            new_name: The new name of the stack component.
        """
        client = Client()

        with console.status(
            f"Renaming {display_name} '{name_id_or_prefix}'...\n"
        ):
            try:
                updated_component = client.update_stack_component(
                    name_id_or_prefix=name_id_or_prefix,
                    component_type=component_type,
                    name=new_name,
                )
            except (KeyError, IllegalOperationError) as err:
                cli_utils.exception(err)

            cli_utils.declare(
                f"Successfully renamed {display_name} `{name_id_or_prefix}` to"
                f" `{new_name}`."
            )
            print_model_url(get_component_url(updated_component))

    return rename_stack_component_command


def generate_stack_component_delete_command(
    component_type: StackComponentType,
) -> Callable[[str], None]:
    """Generates a `delete` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument("name_id_or_prefix", type=str)
    def delete_stack_component_command(name_id_or_prefix: str) -> None:
        """Deletes a stack component.

        Args:
            name_id_or_prefix: The name of the stack component to delete.
        """
        client = Client()

        with console.status(
            f"Deleting {display_name} '{name_id_or_prefix}'...\n"
        ):
            try:
                client.delete_stack_component(
                    name_id_or_prefix=name_id_or_prefix,
                    component_type=component_type,
                )
            except (KeyError, IllegalOperationError) as err:
                cli_utils.exception(err)
            cli_utils.declare(f"Deleted {display_name}: {name_id_or_prefix}")

    return delete_stack_component_command


def generate_stack_component_copy_command(
    component_type: StackComponentType,
) -> Callable[[str, str], None]:
    """Generates a `copy` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "source_component_name_id_or_prefix", type=str, required=True
    )
    @click.argument("target_component", type=str, required=True)
    def copy_stack_component_command(
        source_component_name_id_or_prefix: str,
        target_component: str,
    ) -> None:
        """Copies a stack component.

        Args:
            source_component_name_id_or_prefix: Name or id prefix of the
                                         component to copy.
            target_component: Name of the copied component.
        """
        client = Client()

        with console.status(
            f"Copying {display_name} "
            f"`{source_component_name_id_or_prefix}`..\n"
        ):
            try:
                component_to_copy = client.get_stack_component(
                    name_id_or_prefix=source_component_name_id_or_prefix,
                    component_type=component_type,
                )
            except KeyError as err:
                cli_utils.exception(err)

            copied_component = client.create_stack_component(
                name=target_component,
                flavor=component_to_copy.flavor_name,
                component_type=component_to_copy.type,
                configuration=component_to_copy.configuration,
                labels=component_to_copy.labels,
            )
            print_model_url(get_component_url(copied_component))

    return copy_stack_component_command


def generate_stack_component_logs_command(
    component_type: StackComponentType,
) -> Callable[[str, bool], None]:
    """Generates a `logs` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument("name_id_or_prefix", type=str, required=False)
    @click.option(
        "--follow",
        "-f",
        is_flag=True,
        help="Follow the log file instead of just displaying the current logs.",
    )
    def stack_component_logs_command(
        name_id_or_prefix: str, follow: bool = False
    ) -> None:
        """Displays stack component logs.

        Args:
            name_id_or_prefix: The name of the stack component to display logs
                for.
            follow: Follow the log file instead of just displaying the current
                logs.
        """
        client = Client()

        with console.status(
            f"Fetching the logs for the {display_name} "
            f"'{name_id_or_prefix}'...\n"
        ):
            try:
                component_model = client.get_stack_component(
                    name_id_or_prefix=name_id_or_prefix,
                    component_type=component_type,
                )
            except KeyError as err:
                cli_utils.exception(err)

            from zenml.stack import StackComponent

            component = StackComponent.from_model(
                component_model=component_model
            )
            log_file = component.log_file

            if not log_file or not fileio.exists(log_file):
                cli_utils.warning(
                    f"Unable to find log file for {display_name} "
                    f"'{name_id_or_prefix}'."
                )
                return

        if not log_file or not fileio.exists(log_file):
            cli_utils.warning(
                f"Unable to find log file for {display_name} "
                f"'{component.name}'."
            )
            return

        if follow:
            try:
                with open(log_file, "r") as f:
                    # seek to the end of the file
                    f.seek(0, 2)

                    while True:
                        line = f.readline()
                        if not line:
                            time.sleep(0.1)
                            continue
                        line = line.rstrip("\n")
                        click.echo(line)
            except KeyboardInterrupt:
                cli_utils.declare(f"Stopped following {display_name} logs.")
        else:
            with open(log_file, "r") as f:
                click.echo(f.read())

    return stack_component_logs_command


def generate_stack_component_explain_command(
    component_type: StackComponentType,
) -> Callable[[], None]:
    """Generates an `explain` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """

    def explain_stack_components_command() -> None:
        """Explains the concept of the stack component."""
        component_module = import_module(f"zenml.{component_type.plural}")

        if component_module.__doc__ is not None:
            md = Markdown(component_module.__doc__)
            console.print(md)
        else:
            console.print(
                "The explain subcommand is yet not available for "
                "this stack component. For more information, you can "
                "visit our docs page: https://docs.zenml.io/ and "
                "stay tuned for future releases."
            )

    return explain_stack_components_command


def generate_stack_component_flavor_list_command(
    component_type: StackComponentType,
) -> Callable[[], None]:
    """Generates a `list` command for the flavors of a stack component.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    def list_stack_component_flavor_command() -> None:
        """Lists the flavors for a single type of stack component."""
        client = Client()

        with console.status(f"Listing {display_name} flavors`...\n"):
            flavors = client.get_flavors_by_type(component_type=component_type)

            cli_utils.print_flavor_list(flavors=flavors)
            cli_utils.print_page_info(flavors)

    return list_stack_component_flavor_command


def generate_stack_component_flavor_register_command(
    component_type: StackComponentType,
) -> Callable[[str], None]:
    """Generates a `register` command for the flavors of a stack component.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    command_name = component_type.value.replace("_", "-")
    display_name = _component_display_name(component_type)

    @click.argument(
        "source",
        type=str,
        required=True,
    )
    def register_stack_component_flavor_command(source: str) -> None:
        """Adds a flavor for a stack component type.

        Example:
            Let's say you create an artifact store flavor class `MyArtifactStoreFlavor`
            in the file path `flavors/my_flavor.py`. You would register it as:

            ```shell
            zenml artifact-store flavor register flavors.my_flavor.MyArtifactStoreFlavor
            ```

        Args:
            source: The source path of the flavor class in dot notation format.
        """
        client = Client()

        if not client.root:
            cli_utils.warning(
                f"You're running the `zenml {command_name} flavor register` "
                "command without a ZenML repository. Your current working "
                "directory will be used as the source root relative to which "
                "the `source` argument is expected. To silence this warning, "
                "run `zenml init` at your source code root."
            )

        with console.status(f"Registering a new {display_name} flavor`...\n"):
            try:
                # Register the new model
                new_flavor = client.create_flavor(
                    source=source,
                    component_type=component_type,
                )
            except ValueError as e:
                source_root = source_utils.get_source_root()

                cli_utils.error(
                    f"Flavor registration failed! ZenML tried loading the "
                    f"module `{source}` from path `{source_root}`. If this is "
                    "not what you expect, then please ensure you have run "
                    "`zenml init` at the root of your repository.\n\n"
                    f"Original exception: {str(e)}"
                )

            cli_utils.declare(
                f"Successfully registered new flavor '{new_flavor.name}' "
                f"for stack component '{new_flavor.type}'."
            )

    return register_stack_component_flavor_command


def generate_stack_component_flavor_describe_command(
    component_type: StackComponentType,
) -> Callable[[str], None]:
    """Generates a `describe` command for a single flavor of a component.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "name",
        type=str,
        required=True,
    )
    def describe_stack_component_flavor_command(name: str) -> None:
        """Describes a flavor based on its config schema.

        Args:
            name: The name of the flavor.
        """
        client = Client()

        with console.status(f"Describing {display_name} flavor: {name}`...\n"):
            flavor_model = client.get_flavor_by_name_and_type(
                name=name, component_type=component_type
            )

            cli_utils.describe_pydantic_object(flavor_model.config_schema)
            resources = flavor_model.connector_requirements
            if resources:
                resources_str = f"a '{resources.resource_type}' resource"
                cli_args = f"--resource-type {resources.resource_type}"
                if resources.connector_type:
                    resources_str += (
                        f" provided by a '{resources.connector_type}' "
                        "connector"
                    )
                    cli_args += f"--connector-type {resources.connector_type}"

                cli_utils.declare(
                    f"This flavor supports connecting to external resources "
                    f"with a Service Connector. It requires {resources_str}. "
                    "You can get a list of all available connectors and the "
                    "compatible resources that they can access by running:\n\n"
                    f"'zenml service-connector list-resources {cli_args}'\n"
                    "If no compatible Service Connectors are yet registered, "
                    "you can can register a new one by running:\n\n"
                    f"'zenml service-connector register -i'"
                )
            else:
                cli_utils.declare(
                    "This flavor does not support connecting to external "
                    "resources with a Service Connector."
                )

    return describe_stack_component_flavor_command


def generate_stack_component_flavor_delete_command(
    component_type: StackComponentType,
) -> Callable[[str], None]:
    """Generates a `delete` command for a single flavor of a component.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "name_or_id",
        type=str,
        required=True,
    )
    def delete_stack_component_flavor_command(name_or_id: str) -> None:
        """Deletes a flavor.

        Args:
            name_or_id: The name of the flavor.
        """
        client = Client()

        with console.status(
            f"Deleting a {display_name} flavor: {name_or_id}`...\n"
        ):
            client.delete_flavor(name_or_id)

            cli_utils.declare(f"Successfully deleted flavor '{name_or_id}'.")

    return delete_stack_component_flavor_command


def prompt_select_resource_id(
    resource_ids: List[str],
    resource_name: str,
    interactive: bool = True,
) -> str:
    """Prompts the user to select a resource ID from a list of available IDs.

    Args:
        resource_ids: A list of available resource IDs.
        resource_name: The name of the resource type to select.
        interactive: Whether to prompt the user for input or error out if
            user input is required.

    Returns:
        The selected resource ID.
    """
    if len(resource_ids) == 1:
        # Only one resource ID is available, so we can select it
        # without prompting the user
        return resource_ids[0]

    if len(resource_ids) > 1:
        resource_ids_list = "\n - " + "\n - ".join(resource_ids)
        msg = (
            f"Multiple {resource_name} resources are available for the "
            f"selected connector:\n{resource_ids_list}\n"
        )
        # User needs to select a resource ID from the list
        if not interactive:
            cli_utils.error(
                f"{msg}Please use the `--resource-id` command line "  # nosec
                f"argument to select a {resource_name} resource from the "
                "list."
            )
        resource_id = click.prompt(
            f"{msg}Please select the {resource_name} that you want to use",
            type=click.Choice(resource_ids),
            show_choices=False,
        )

        return cast(str, resource_id)

    # We should never get here, but just in case...
    cli_utils.error(
        "Could not determine which resource to use. Please select a "
        "different connector."
    )


def prompt_select_resource(
    resource_list: List[ServiceConnectorResourcesModel],
) -> Tuple[UUID, str]:
    """Prompts the user to select a resource ID from a list of resources.

    Args:
        resource_list: List of resources to select from.

    Returns:
        The ID of a selected connector and the ID of the selected resource
        instance.
    """
    if len(resource_list) == 1:
        click.echo("Only one connector has compatible resources:")
    else:
        click.echo("The following connectors have compatible resources:")

    cli_utils.print_service_connector_resource_table(resource_list)

    if len(resource_list) == 1:
        connect = click.confirm(
            "Would you like to use this connector?",
            default=True,
        )
        if not connect:
            cli_utils.error("Aborting.")
        resources = resource_list[0]
    else:
        # Prompt the user to select a connector by its name or ID
        while True:
            connector_id = click.prompt(
                "Please enter the name or ID of the connector you want to use",
                type=click.Choice(
                    [
                        str(connector.id)
                        for connector in resource_list
                        if connector.id is not None
                    ]
                    + [
                        connector.name
                        for connector in resource_list
                        if connector.name is not None
                    ]
                ),
                show_choices=False,
            )
            matches = [
                c
                for c in resource_list
                if str(c.id) == connector_id or c.name == connector_id
            ]
            if len(matches) > 1:
                cli_utils.declare(
                    f"Multiple connectors with name '{connector_id}' "
                    "were found. Please try again."
                )
            else:
                resources = matches[0]
                break

    connector_uuid = resources.id
    assert connector_uuid is not None

    assert len(resources.resources) == 1
    resource_name = resources.resources[0].resource_type
    if not isinstance(resources.connector_type, str):
        resource_type_spec = resources.connector_type.resource_type_dict[
            resource_name
        ]
        resource_name = resource_type_spec.name

    resource_id = prompt_select_resource_id(
        resources.resources[0].resource_ids or [], resource_name=resource_name
    )

    return connector_uuid, resource_id


def generate_stack_component_connect_command(
    component_type: StackComponentType,
) -> Callable[[str, str], None]:
    """Generates a `connect` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    _component_display_name(component_type)

    @click.argument(
        "name_id_or_prefix",
        type=str,
        required=False,
    )
    @click.option(
        "--connector",
        "-c",
        "connector",
        help="The name, ID or prefix of the connector to use.",
        required=False,
        type=str,
    )
    @click.option(
        "--resource-id",
        "-r",
        "resource_id",
        help="The resource ID to use with the connector. Only required for "
        "multi-instance connectors that are not already configured with a "
        "particular resource ID.",
        required=False,
        type=str,
    )
    @click.option(
        "--interactive",
        "-i",
        "interactive",
        is_flag=True,
        default=False,
        help="Configure a service connector resource interactively.",
        type=click.BOOL,
    )
    @click.option(
        "--no-verify",
        "no_verify",
        is_flag=True,
        default=False,
        help="Skip verification of the connector resource.",
        type=click.BOOL,
    )
    def connect_stack_component_command(
        name_id_or_prefix: Optional[str],
        connector: Optional[str] = None,
        resource_id: Optional[str] = None,
        interactive: bool = False,
        no_verify: bool = False,
    ) -> None:
        """Connect the stack component to a resource through a service connector.

        Args:
            name_id_or_prefix: The name of the stack component to connect.
            connector: The name, ID or prefix of the connector to use.
            resource_id: The resource ID to use connect to. Only
                required for multi-instance connectors that are not already
                configured with a particular resource ID.
            interactive: Configure a service connector resource interactively.
            no_verify: Do not verify whether the resource is accessible.
        """
        connect_stack_component_with_service_connector(
            component_type=component_type,
            name_id_or_prefix=name_id_or_prefix,
            connector=connector,
            resource_id=resource_id,
            interactive=interactive,
            no_verify=no_verify,
        )

    return connect_stack_component_command


def generate_stack_component_disconnect_command(
    component_type: StackComponentType,
) -> Callable[[str], None]:
    """Generates a `disconnect` command for the specific stack component type.

    Args:
        component_type: Type of the component to generate the command for.

    Returns:
        A function that can be used as a `click` command.
    """
    display_name = _component_display_name(component_type)

    @click.argument(
        "name_id_or_prefix",
        type=str,
        required=True,
    )
    def disconnect_stack_component_command(name_id_or_prefix: str) -> None:
        """Disconnect a stack component from a service connector.

        Args:
            name_id_or_prefix: The name of the stack component to disconnect.
        """
        client = Client()

        with console.status(
            f"Disconnecting service-connector from {display_name} '{name_id_or_prefix}'...\n"
        ):
            try:
                updated_component = client.update_stack_component(
                    name_id_or_prefix=name_id_or_prefix,
                    component_type=component_type,
                    disconnect=True,
                )
            except (KeyError, IllegalOperationError) as err:
                cli_utils.exception(err)

            cli_utils.declare(
                f"Successfully disconnected the service-connector from {display_name} `{name_id_or_prefix}`."
            )
            print_model_url(get_component_url(updated_component))

    return disconnect_stack_component_command


def register_single_stack_component_cli_commands(
    component_type: StackComponentType, parent_group: click.Group
) -> None:
    """Registers all basic stack component CLI commands.

    Args:
        component_type: Type of the component to generate the command for.
        parent_group: The parent group to register the commands to.
    """
    command_name = component_type.value.replace("_", "-")
    singular_display_name = _component_display_name(component_type)
    plural_display_name = _component_display_name(component_type, plural=True)

    @parent_group.group(
        command_name,
        cls=TagGroup,
        help=f"Commands to interact with {plural_display_name}.",
        tag=CliCategories.STACK_COMPONENTS,
    )
    def command_group() -> None:
        """Group commands for a single stack component type."""

    # zenml stack-component get
    get_command = generate_stack_component_get_command(component_type)
    command_group.command(
        "get", help=f"Get the name of the active {singular_display_name}."
    )(get_command)

    # zenml stack-component describe
    describe_command = generate_stack_component_describe_command(
        component_type
    )
    command_group.command(
        "describe",
        help=f"Show details about the (active) {singular_display_name}.",
    )(describe_command)

    # zenml stack-component list
    list_command = generate_stack_component_list_command(component_type)
    command_group.command(
        "list", help=f"List all registered {plural_display_name}."
    )(list_command)

    # zenml stack-component register
    register_command = generate_stack_component_register_command(
        component_type
    )
    context_settings = {"ignore_unknown_options": True}
    command_group.command(
        "register",
        context_settings=context_settings,
        help=f"Register a new {singular_display_name}.",
    )(register_command)

    # zenml stack-component update
    update_command = generate_stack_component_update_command(component_type)
    context_settings = {"ignore_unknown_options": True}
    command_group.command(
        "update",
        context_settings=context_settings,
        help=f"Update a registered {singular_display_name}.",
    )(update_command)

    # zenml stack-component remove-attribute
    remove_attribute_command = (
        generate_stack_component_remove_attribute_command(component_type)
    )
    context_settings = {"ignore_unknown_options": True}
    command_group.command(
        "remove-attribute",
        context_settings=context_settings,
        help=f"Remove attributes from a registered {singular_display_name}.",
    )(remove_attribute_command)

    # zenml stack-component rename
    rename_command = generate_stack_component_rename_command(component_type)
    command_group.command(
        "rename", help=f"Rename a registered {singular_display_name}."
    )(rename_command)

    # zenml stack-component delete
    delete_command = generate_stack_component_delete_command(component_type)
    command_group.command(
        "delete", help=f"Delete a registered {singular_display_name}."
    )(delete_command)

    # zenml stack-component copy
    copy_command = generate_stack_component_copy_command(component_type)
    command_group.command(
        "copy", help=f"Copy a registered {singular_display_name}."
    )(copy_command)

    # zenml stack-component logs
    logs_command = generate_stack_component_logs_command(component_type)
    command_group.command(
        "logs", help=f"Display {singular_display_name} logs."
    )(logs_command)

    # zenml stack-component connect
    connect_command = generate_stack_component_connect_command(component_type)
    command_group.command(
        "connect",
        help=f"Connect {singular_display_name} to a service connector.",
    )(connect_command)

    # zenml stack-component connect
    disconnect_command = generate_stack_component_disconnect_command(
        component_type
    )
    command_group.command(
        "disconnect",
        help=f"Disconnect {singular_display_name} from a service connector.",
    )(disconnect_command)

    # zenml stack-component explain
    explain_command = generate_stack_component_explain_command(component_type)
    command_group.command(
        "explain", help=f"Explaining the {plural_display_name}."
    )(explain_command)

    # zenml stack-component flavor
    @command_group.group(
        "flavor", help=f"Commands to interact with {plural_display_name}."
    )
    def flavor_group() -> None:
        """Group commands to handle flavors for a stack component type."""

    # zenml stack-component flavor register
    register_flavor_command = generate_stack_component_flavor_register_command(
        component_type=component_type
    )
    flavor_group.command(
        "register",
        help=f"Register a new {singular_display_name} flavor.",
    )(register_flavor_command)

    # zenml stack-component flavor list
    list_flavor_command = generate_stack_component_flavor_list_command(
        component_type=component_type
    )
    flavor_group.command(
        "list",
        help=f"List all registered flavors for {plural_display_name}.",
    )(list_flavor_command)

    # zenml stack-component flavor describe
    describe_flavor_command = generate_stack_component_flavor_describe_command(
        component_type=component_type
    )
    flavor_group.command(
        "describe",
        help=f"Describe a {singular_display_name} flavor.",
    )(describe_flavor_command)

    # zenml stack-component flavor delete
    delete_flavor_command = generate_stack_component_flavor_delete_command(
        component_type=component_type
    )
    flavor_group.command(
        "delete",
        help=f"Delete a {plural_display_name} flavor.",
    )(delete_flavor_command)


def register_all_stack_component_cli_commands() -> None:
    """Registers CLI commands for all stack components."""
    for component_type in StackComponentType:
        register_single_stack_component_cli_commands(
            component_type, parent_group=cli
        )


def connect_stack_component_with_service_connector(
    component_type: StackComponentType,
    name_id_or_prefix: Optional[str] = None,
    connector: Optional[str] = None,
    resource_id: Optional[str] = None,
    interactive: bool = False,
    no_verify: bool = False,
) -> None:
    """Connect the stack component to a resource through a service connector.

    Args:
        component_type: Type of the component to generate the command for.
        name_id_or_prefix: The name of the stack component to connect.
        connector: The name, ID or prefix of the connector to use.
        resource_id: The resource ID to use connect to. Only
            required for multi-instance connectors that are not already
            configured with a particular resource ID.
        interactive: Configure a service connector resource interactively.
        no_verify: Do not verify whether the resource is accessible.
    """
    display_name = _component_display_name(component_type)

    if not connector and not interactive:
        cli_utils.error(
            "Please provide either a connector ID or set the interactive flag."
        )

    if connector and interactive:
        cli_utils.error(
            "Please provide either a connector ID or set the interactive "
            "flag, not both."
        )

    client = Client()

    try:
        component_model = client.get_stack_component(
            name_id_or_prefix=name_id_or_prefix,
            component_type=component_type,
        )
    except KeyError as err:
        cli_utils.exception(err)

    requirements = component_model.flavor.connector_requirements

    if not requirements:
        cli_utils.error(
            f"The '{component_model.name}' {display_name} implementation "
            "does not support using a service connector to connect to "
            "resources."
        )

    resource_type = requirements.resource_type
    if requirements.resource_id_attr is not None:
        # Check if an attribute is set in the component configuration
        resource_id = component_model.configuration.get(
            requirements.resource_id_attr
        )

    if interactive:
        # Fetch the list of connectors that have resources compatible with
        # the stack component's flavor's resource requirements
        with console.status(
            "Finding all resources matching the stack component "
            "requirements (this could take a while)...\n"
        ):
            resource_list = client.list_service_connector_resources(
                connector_type=requirements.connector_type,
                resource_type=resource_type,
                resource_id=resource_id,
            )

        resource_list = [
            resource
            for resource in resource_list
            if resource.resources[0].resource_ids
        ]

        error_resource_list = [
            resource
            for resource in resource_list
            if not resource.resources[0].resource_ids
        ]

        if not resource_list:
            # No compatible resources were found
            additional_info = ""
            if error_resource_list:
                additional_info = (
                    f"{len(error_resource_list)} connectors can be used "
                    f"to gain access to {resource_type} resources required "
                    "for the stack component, but they are in an error "
                    "state or they didn't list any matching resources. "
                )
            command_args = ""
            if requirements.connector_type:
                command_args += (
                    f" --connector-type {requirements.connector_type}"
                )
            command_args += f" --resource-type {requirements.resource_type}"
            if resource_id:
                command_args += f" --resource-id {resource_id}"

            cli_utils.error(
                f"No compatible valid resources were found for the "
                f"'{component_model.name}' {display_name}. "
                f"{additional_info}You can create a new "
                "connector using the 'zenml service-connector register' "
                "command or list the compatible resources using the "
                f"'zenml service-connector list-resources{command_args}' "
                "command."
            )

        # Prompt the user to select a connector and a resource ID, if
        # applicable
        connector_id, resource_id = prompt_select_resource(resource_list)
        no_verify = False
    else:
        # Non-interactive mode: we need to fetch the connector model first

        assert connector is not None
        try:
            connector_model = client.get_service_connector(connector)
        except KeyError as err:
            cli_utils.error(
                f"Could not find a connector '{connector}': {str(err)}"
            )

        connector_id = connector_model.id

        satisfied, msg = requirements.is_satisfied_by(
            connector_model, component_model
        )
        if not satisfied:
            cli_utils.error(
                f"The connector with ID {connector_id} does not match the "
                f"component's `{name_id_or_prefix}` of type `{component_type}`"
                f" connector requirements: {msg}. Please pick a connector that "
                f"is compatible with the component flavor and try again, or "
                f"use the interactive mode to select a compatible connector."
            )

        if not resource_id:
            if connector_model.resource_id:
                resource_id = connector_model.resource_id
            elif connector_model.supports_instances:
                cli_utils.error(
                    f"Multiple {resource_type} resources are available for "
                    "the selected connector. Please use the "
                    "`--resource-id` command line argument to configure a "
                    f"{resource_type} resource or use the interactive mode "
                    "to select a resource interactively."
                )

    connector_resources: Optional[ServiceConnectorResourcesModel] = None
    if not no_verify:
        with console.status(
            "Validating service connector resource configuration...\n"
        ):
            try:
                connector_resources = client.verify_service_connector(
                    connector_id,
                    resource_type=requirements.resource_type,
                    resource_id=resource_id,
                )
            except (
                KeyError,
                ValueError,
                IllegalOperationError,
                NotImplementedError,
                AuthorizationException,
            ) as e:
                cli_utils.error(
                    f"Access to the resource could not be verified: {e}"
                )
        resources = connector_resources.resources[0]
        if resources.resource_ids:
            if len(resources.resource_ids) > 1:
                cli_utils.error(
                    f"Multiple {resource_type} resources are available for "
                    "the selected connector. Please use the "
                    "`--resource-id` command line argument to configure a "
                    f"{resource_type} resource or use the interactive mode "
                    "to select a resource interactively."
                )
            else:
                resource_id = resources.resource_ids[0]

    with console.status(f"Updating {display_name} '{name_id_or_prefix}'...\n"):
        try:
            client.update_stack_component(
                name_id_or_prefix=name_id_or_prefix,
                component_type=component_type,
                connector_id=connector_id,
                connector_resource_id=resource_id,
            )
        except (KeyError, IllegalOperationError) as err:
            cli_utils.exception(err)

    if connector_resources is not None:
        cli_utils.declare(
            f"Successfully connected {display_name} "
            f"`{component_model.name}` to the following resources:"
        )

        cli_utils.print_service_connector_resource_table([connector_resources])

    else:
        cli_utils.declare(
            f"Successfully connected {display_name} "
            f"`{component_model.name}` to resource."
        )


register_all_stack_component_cli_commands()
register_annotator_subcommands()
register_feature_store_subcommands()
register_model_deployer_subcommands()
register_model_registry_subcommands()



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Module for centralized WarningController implementation."""

import logging
from collections import Counter
from typing import Any

from zenml.enums import LoggingLevels
from zenml.utils.singleton import SingletonMetaClass
from zenml.utils.warnings.base import WarningConfig, WarningVerbosity
from zenml.utils.warnings.registry import WarningCodes

logger = logging.getLogger(__name__)


class WarningController(metaclass=SingletonMetaClass):
    """Class responsible for centralized handling of warning messages."""

    def __init__(self) -> None:
        """WarningController constructor."""
        self._warning_configs: dict[str, WarningConfig] = {}
        self._warning_statistics: dict[str, int] = Counter()

    def register(self, warning_configs: dict[str, WarningConfig]) -> None:
        """Register a warning config collection to the controller.

        Args:
            warning_configs: Configs to be registered. Key should be the warning code.
        """
        self._warning_configs.update(warning_configs)

    @staticmethod
    def _resolve_call_details() -> tuple[str, int]:
        import inspect

        frame = inspect.stack()[3]  # public methods -> _log -> _resolve
        module = inspect.getmodule(frame[0])
        module_name = module.__name__ if module else "<unknown module>"
        line_number = frame.lineno

        return module_name, line_number

    @staticmethod
    def _get_display_message(
        message: str,
        module_name: str,
        line_number: int,
        config: WarningConfig,
    ) -> str:
        """Helper method to build the warning message string.

        Args:
            message: The warning message.
            module_name: The module that the warning call originated from.
            line_number: The line number that the warning call originated from.
            config: The warning configuration.

        Returns:
            A warning message containing extra fields/info based on warning config.
        """
        display = f"[{config.code}]({config.category}) - {message}"

        if config.verbosity == WarningVerbosity.MEDIUM:
            display = f"{module_name}:{line_number} {display}"

        if config.verbosity == WarningVerbosity.HIGH:
            display = f"{display}\n{config.description}"

        return display

    def _log(
        self,
        warning_code: WarningCodes,
        message: str,
        level: LoggingLevels,
        **kwargs: dict[str, Any],
    ) -> None:
        """Core function for warning handling.

        Args:
            warning_code: The code of the warning configuration.
            message: The warning message.
            level: The level of the warning.
            **kwargs: Keyword arguments (for formatted messages).

        """
        warning_config = self._warning_configs.get(warning_code)

        # resolves the module and line number of the warning call.
        module_name, line_number = self._resolve_call_details()

        if not warning_config:
            # If no config is available just follow default behavior:
            logger.warning(f"{module_name}:{line_number} - {message}")
            return

        if warning_config.is_throttled:
            if warning_code in self._warning_statistics:
                # Throttled code has already appeared - skip.
                return

        display_message = self._get_display_message(
            message=message,
            module_name=module_name,
            line_number=line_number,
            config=warning_config,
        )

        self._warning_statistics[warning_code] += 1

        if level == LoggingLevels.INFO:
            logger.info(display_message.format(**kwargs))
        else:
            # Assumes warning level is the default if an invalid option is passed.
            logger.warning(display_message.format(**kwargs))

    def warn(
        self, *, warning_code: WarningCodes, message: str, **kwargs: Any
    ) -> None:
        """Method to execute warning handling logic with warning log level.

        Args:
            warning_code: The code of the warning (see WarningCodes enum)
            message: The message to display.
            **kwargs: Keyword arguments (for formatted messages).
        """
        self._log(warning_code, message, LoggingLevels.WARNING, **kwargs)

    def info(
        self, *, warning_code: WarningCodes, message: str, **kwargs: Any
    ) -> None:
        """Method to execute warning handling logic with info log level.

        Args:
            warning_code: The code of the warning (see WarningCodes enum)
            message: The message to display.
            **kwargs: Keyword arguments (for formatted messages).
        """
        self._log(warning_code, message, LoggingLevels.INFO, **kwargs)

    @staticmethod
    def create() -> "WarningController":
        """Factory function for WarningController.

        Creates a new warning controller and registers system warning configs.

        Returns:
            A warning controller instance.
        """
        from zenml.utils.warnings.registry import WARNING_CONFIG_REGISTRY

        registry = WarningController()

        registry.register(warning_configs=WARNING_CONFIG_REGISTRY)

        return registry



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""The analytics module of ZenML.

This module is based on the 'analytics-python' package created by Segment.
The base functionalities are adapted to work with the ZenML analytics server.
"""

import locale
from types import TracebackType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union
from uuid import UUID

from zenml import __version__
from zenml.analytics.client import default_client
from zenml.analytics.utils import is_analytics_disabled_internally
from zenml.constants import (
    ENV_ZENML_SERVER,
    handle_bool_env_var,
)
from zenml.environment import Environment, get_environment
from zenml.logger import get_logger
from zenml.utils.time_utils import utc_now_tz_aware

if TYPE_CHECKING:
    from zenml.analytics.enums import AnalyticsEvent
    from zenml.models import (
        ServerDatabaseType,
        ServerDeploymentType,
    )

Json = Union[Dict[str, Any], List[Any], str, int, float, bool, None]

logger = get_logger(__name__)


class AnalyticsContext:
    """Client class for ZenML Analytics v2."""

    def __init__(self) -> None:
        """Initialization.

        Use this as a context manager to ensure that analytics are initialized
        properly, only tracked when configured to do so and that any errors
        are handled gracefully.
        """
        self.analytics_opt_in: bool = False

        self.user_id: Optional[UUID] = None
        self.external_user_id: Optional[UUID] = None
        self.executed_by_service_account: Optional[bool] = None
        self.client_id: Optional[UUID] = None
        self.server_id: Optional[UUID] = None
        self.external_server_id: Optional[UUID] = None
        self.server_metadata: Optional[Dict[str, str]] = None

        self.database_type: Optional["ServerDatabaseType"] = None
        self.deployment_type: Optional["ServerDeploymentType"] = None

    def __enter__(self) -> "AnalyticsContext":
        """Enter analytics context manager.

        Returns:
            The analytics context.
        """
        if is_analytics_disabled_internally():
            self.analytics_opt_in = False
            return self

        # Fetch the analytics opt-in setting
        from zenml.config.global_config import GlobalConfiguration

        try:
            gc = GlobalConfiguration()

            if not gc.is_initialized:
                # If the global configuration is not initialized, using the
                # zen store can lead to multiple initialization issues, because
                # the analytics are triggered during the initialization of the
                # zen store.
                return self

            store_info = gc.zen_store.get_store_info()

            # For local ZenML servers, we always use the client's analytics
            # opt-in configuration.
            if self.in_server and not store_info.is_local():
                self.analytics_opt_in = store_info.analytics_enabled
            else:
                self.analytics_opt_in = gc.analytics_opt_in

            if not self.analytics_opt_in:
                return self

            # Fetch the `user_id`
            if self.in_server:
                from zenml.zen_server.utils import (
                    get_auth_context,
                    server_config,
                )

                # If the code is running on the server, use the auth context.
                auth_context = get_auth_context()
                if auth_context is not None:
                    self.user_id = auth_context.user.id
                    self.executed_by_service_account = (
                        auth_context.user.is_service_account
                    )
                    self.external_user_id = auth_context.user.external_user_id

                self.external_server_id = server_config().external_server_id
                self.project_id = None
            else:
                from zenml.client import Client

                # If the code is running on the client, use the default user.
                active_user = gc.zen_store.get_user()
                self.user_id = active_user.id
                self.executed_by_service_account = (
                    active_user.is_service_account
                )
                self.external_user_id = active_user.external_user_id
                self.project_id = Client().active_project.id

            # Fetch the `client_id`
            if self.in_server:
                # If the code is running on the server, there is no client id.
                self.client_id = None
            else:
                # If the code is running on the client, attach the client id.
                self.client_id = gc.user_id

            self.server_id = store_info.id
            self.deployment_type = store_info.deployment_type
            self.database_type = store_info.database_type
            self.server_metadata = store_info.metadata
        except Exception as e:
            self.analytics_opt_in = False
            logger.debug(f"Analytics initialization failed: {e}")

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """Exit context manager.

        Args:
            exc_type: Exception type.
            exc_val: Exception value.
            exc_tb: Exception traceback.

        Returns:
            True.
        """
        if exc_val is not None:
            logger.debug(f"Sending telemetry data failed: {exc_val}")

        return True

    @property
    def in_server(self) -> bool:
        """Flag to check whether the code is running in a ZenML server.

        Returns:
            True if running in a server, False otherwise.
        """
        return handle_bool_env_var(ENV_ZENML_SERVER)

    def identify(self, traits: Optional[Dict[str, Any]] = None) -> bool:
        """Identify the user through segment.

        Args:
            traits: Traits of the user.

        Returns:
            True if tracking information was sent, False otherwise.
        """
        success = False
        if self.analytics_opt_in and self.user_id is not None:
            success, _ = default_client.identify(
                user_id=self.user_id,
                traits=traits,
            )

        return success

    def alias(self, user_id: UUID, previous_id: UUID) -> bool:
        """Alias user IDs.

        Args:
            user_id: The user ID.
            previous_id: Previous ID for the alias.

        Returns:
            True if alias information was sent, False otherwise.
        """
        success = False
        if self.analytics_opt_in:
            success, _ = default_client.alias(
                user_id=user_id,
                previous_id=previous_id,
            )

        return success

    def group(
        self,
        group_id: UUID,
        traits: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """Group the user.

        Args:
            group_id: Group ID.
            traits: Traits of the group.

        Returns:
            True if tracking information was sent, False otherwise.
        """
        success = False
        if self.analytics_opt_in and self.user_id is not None:
            success, _ = default_client.group(
                user_id=self.user_id,
                group_id=group_id,
                traits=traits or {},
            )

        return success

    def track(
        self,
        event: "Union[AnalyticsEvent, str]",
        properties: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """Track an event.

        Args:
            event: Event to track. Can be an AnalyticsEvent enum member
                or an arbitrary string for downstream consumers that
                define their own event names.
            properties: Event properties.

        Returns:
            True if tracking information was sent, False otherwise.
        """
        from zenml.analytics.enums import AnalyticsEvent

        if properties is None:
            properties = {}

        event_value = (
            event.value if isinstance(event, AnalyticsEvent) else event
        )
        if (
            not self.analytics_opt_in
            and event_value
            not in {
                AnalyticsEvent.OPT_OUT_ANALYTICS.value,
                AnalyticsEvent.OPT_IN_ANALYTICS.value,
            }
            or self.user_id is None
        ):
            return False

        # add basics
        properties.update(Environment.get_system_info())
        properties.update(
            {
                "environment": get_environment(),
                "python_version": Environment.python_version(),
                "version": __version__,
                "client_id": str(self.client_id),
                "user_id": str(self.user_id),
                "server_id": str(self.server_id),
                "deployment_type": str(self.deployment_type),
                "database_type": str(self.database_type),
                "executed_by_service_account": self.executed_by_service_account,
            }
        )

        try:
            # Timezone as tzdata
            tz = utc_now_tz_aware().astimezone().tzname()
            if tz is not None:
                properties.update({"timezone": tz})

            # Language code such as "en_DE"
            language_code, encoding = locale.getlocale()
            if language_code is not None:
                properties.update({"locale": language_code})
        except Exception:
            pass

        if self.external_user_id:
            properties["external_user_id"] = self.external_user_id

        if self.external_server_id:
            properties["external_server_id"] = self.external_server_id

        if self.server_metadata:
            properties.update(self.server_metadata)

        if self.project_id:
            properties.setdefault("project_id", str(self.project_id))

        for k, v in properties.items():
            if isinstance(v, UUID):
                properties[k] = str(v)

        success, _ = default_client.track(
            user_id=self.user_id,
            event=event,
            properties=properties,
        )

        logger.debug(
            f"Sending analytics: User: {self.user_id}, Event: {event}, "
            f"Metadata: {properties}"
        )

        return success



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Custom types that can be used as metadata of ZenML artifacts."""

import json
from typing import Any, Dict, List, Set, Tuple, Union

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

from zenml.constants import STR_FIELD_MAX_LENGTH, TEXT_FIELD_MAX_LENGTH
from zenml.logger import get_logger
from zenml.utils.enum_utils import StrEnum
from zenml.utils.json_utils import pydantic_encoder

logger = get_logger(__name__)


class Uri(str):
    """Special string class to indicate a URI."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Additional method for pydantic to recognize it as a valid type.

        Args:
            source_type: the source type
            handler: the handler

        Returns:
            the schema for the custom type.
        """
        return core_schema.no_info_after_validator_function(cls, handler(str))


class Path(str):
    """Special string class to indicate a path."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Additional method for pydantic to recognize it as a valid type.

        Args:
            source_type: the source type
            handler: the handler

        Returns:
            the schema for the custom type.
        """
        return core_schema.no_info_after_validator_function(cls, handler(str))


class DType(str):
    """Special string class to indicate a data type."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Additional method for pydantic to recognize it as a valid type.

        Args:
            source_type: the source type
            handler: the handler

        Returns:
            the schema for the custom type.
        """
        return core_schema.no_info_after_validator_function(cls, handler(str))


class StorageSize(int):
    """Special int class to indicate the storage size in number of bytes."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Additional method for pydantic to recognize it as a valid type.

        Args:
            source_type: the source type
            handler: the handler

        Returns:
            the schema for the custom type.
        """
        return core_schema.no_info_after_validator_function(cls, handler(int))


# Union of all types that can be used as metadata. Used for type annotations.
MetadataType = Union[
    str,
    int,
    float,
    bool,
    Dict[Any, Any],
    List[Any],
    Set[Any],
    Tuple[Any, ...],
    Uri,
    Path,
    DType,
    StorageSize,
]

# Tuple of all types that can be used as metadata. Used for `isinstance` checks.
MetadataTypeTuple = (
    str,
    int,
    float,
    bool,
    dict,
    list,
    set,
    tuple,
    Uri,
    Path,
    DType,
    StorageSize,
)


class MetadataTypeEnum(StrEnum):
    """String Enum of all possible types that metadata can have."""

    STRING = "str"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    LIST = "list"
    DICT = "dict"
    TUPLE = "tuple"
    SET = "set"
    URI = "Uri"
    PATH = "Path"
    DTYPE = "DType"
    STORAGE_SIZE = "StorageSize"


metadata_type_to_enum_mapping = {
    str: MetadataTypeEnum.STRING,
    int: MetadataTypeEnum.INT,
    float: MetadataTypeEnum.FLOAT,
    bool: MetadataTypeEnum.BOOL,
    dict: MetadataTypeEnum.DICT,
    list: MetadataTypeEnum.LIST,
    set: MetadataTypeEnum.SET,
    tuple: MetadataTypeEnum.TUPLE,
    Uri: MetadataTypeEnum.URI,
    Path: MetadataTypeEnum.PATH,
    DType: MetadataTypeEnum.DTYPE,
    StorageSize: MetadataTypeEnum.STORAGE_SIZE,
}

metadata_enum_to_type_mapping = {
    value: key for key, value in metadata_type_to_enum_mapping.items()
}


def get_metadata_type(
    object_: object,
) -> MetadataTypeEnum:
    """Get the metadata type enum for a given object.

    Args:
        object_: The object to get the metadata type for.

    Returns:
        The corresponding metadata type enum.

    Raises:
        ValueError: If the metadata type is not supported.
    """
    metadata_type = type(object_)
    if metadata_type in metadata_type_to_enum_mapping:
        return metadata_type_to_enum_mapping[metadata_type]
    raise ValueError(f"Metadata type {metadata_type} is not supported.")


def cast_to_metadata_type(
    value: object,
    type_: MetadataTypeEnum,
) -> MetadataType:
    """Cast an object to a metadata type.

    Args:
        value: The object to cast.
        type_: The metadata type to cast to.

    Returns:
        The value cast to the given metadata type.
    """
    metadata_type = metadata_enum_to_type_mapping[type_]
    typed_value = metadata_type(value)
    return typed_value  # type: ignore[no-any-return]


def validate_metadata(
    metadata: Dict[str, MetadataType],
) -> Dict[str, MetadataType]:
    """Validate metadata.

    This function excludes and warns about metadata values that are too long
    or of an unsupported type.

    Args:
        metadata: The metadata to validate.

    Returns:
        The validated metadata.
    """
    validated_metadata = {}

    for key, value in metadata.items():
        if len(key) > STR_FIELD_MAX_LENGTH:
            logger.warning(
                f"Metadata key '{key}' is too large to be "
                "stored in the database. Skipping."
            )
            continue

        if (
            len(json.dumps(value, default=pydantic_encoder))
            > TEXT_FIELD_MAX_LENGTH
        ):
            logger.warning(
                f"Metadata value for key '{key}' is too large to be "
                "stored in the database. Skipping."
            )
            continue

        try:
            get_metadata_type(value)
        except ValueError as e:
            logger.warning(
                f"Metadata value for key '{key}' is not of a supported "
                f"type. Skipping. Full error: {e}"
            )

        validated_metadata[key] = value

    return validated_metadata



# Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQL Model Implementations for Triggers."""

import json
from datetime import datetime
from typing import Any, Sequence
from uuid import UUID

from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import selectinload
from sqlalchemy.orm.session import object_session
from sqlalchemy.sql.base import ExecutableOption
from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.sqltypes import TEXT, VARCHAR, String
from sqlmodel import Field, Relationship, col, desc, select

from zenml.constants import TEXT_FIELD_MAX_LENGTH
from zenml.enums import TriggerFlavor, TriggerType
from zenml.models import (
    TRIGGER_RETURN_TYPE_UNION,
    TriggerRequest,
    TriggerResponseMetadata,
    TriggerResponseResources,
    TriggerUpdate,
)
from zenml.triggers.registry import (
    TYPE_TO_RESPONSE_BODY_MAPPING,
    TYPE_TO_RESPONSE_MAPPING,
)
from zenml.zen_stores.schemas import PipelineRunSchema, PipelineSnapshotSchema
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
    build_index,
)
from zenml.zen_stores.schemas.trigger_assoc import (
    TriggerExecutionSchema,
    TriggerSnapshotSchema,
)
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import (
    jl_arg,
)


class TriggerSchema(NamedSchema, table=True):
    """SQL Model for schedules."""

    __tablename__ = "trigger"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "project_id",
            name="unique_trigger_name_in_project",
        ),
        build_index(
            table_name=__tablename__,
            column_names=[
                "type",
            ],
        ),
        build_index(
            table_name=__tablename__,
            column_names=["flavor", "next_occurrence"],
        ),
    )

    # -------------------- BASE FIELDS

    active: bool = Field(
        nullable=False, description="Whether the trigger is active."
    )
    is_archived: bool = Field(
        nullable=False,
        default=False,
        description="Soft-deletion flag. If Trigger is set to archived"
        " it is preserved in the DB but otherwise unusable.",
    )
    type: str = Field(
        sa_column=Column(VARCHAR(20), nullable=False),
        description="The Trigger type e.g. webhook, schedule, etc.",
    )
    flavor: str = Field(
        sa_column=Column(VARCHAR(50), nullable=False),
        description="The Trigger category e.g. native schedule, github webhook, etc.",
    )
    configuration: str = Field(
        sa_column=Column(
            String(length=TEXT_FIELD_MAX_LENGTH).with_variant(TEXT, "mysql"),
            nullable=False,
        ),
        description="JSON object - the extra trigger object parameters",
    )
    concurrency: str = Field(
        sa_column=Column(VARCHAR(20), nullable=False),
        description="The trigger execution concurrency (SKIP, SUBMIT, etc.)",
    )

    # -------------------- FOREIGN KEYS ------------------------------------

    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )

    project: ProjectSchema = Relationship(back_populates="triggers")

    user_id: UUID | None = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    # ------------------ RELATIONSHIPS --------------------------------------------

    snapshots: list[PipelineSnapshotSchema] = Relationship(
        link_model=TriggerSnapshotSchema,
        sa_relationship_kwargs={"viewonly": True},
    )

    user: UserSchema | None = Relationship(back_populates="triggers")

    # ------------------ FLAT DATA FIELDS FOR FAST FILTERING ----------------------

    next_occurrence: datetime | None = Field(
        nullable=True,
        default=None,
        description="The next occurrence. Applicable for schedules.",
    )

    @property
    def latest_run(self) -> PipelineRunSchema | None:
        """Fetch the latest execution for this trigger.

        Raises:
            RuntimeError: If no session for the schema exists.

        Returns:
            The latest run for this pipeline.
        """
        if session := object_session(self):
            stmt = (
                select(PipelineRunSchema)
                .join(
                    TriggerExecutionSchema,
                    col(TriggerExecutionSchema.pipeline_run_id)
                    == col(PipelineRunSchema.id),
                )
                .where(TriggerExecutionSchema.trigger_id == self.id)
                .order_by(desc(TriggerExecutionSchema.created_at))
                .limit(1)
            )

            return session.execute(stmt).scalars().one_or_none()
        else:
            raise RuntimeError(
                "Missing DB session to fetch latest pipeline run for trigger."
            )

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        if include_resources:
            options.extend(
                [
                    selectinload(jl_arg(TriggerSchema.snapshots)),
                ]
            )

        return options

    @classmethod
    def from_request(cls, trigger_request: TriggerRequest) -> "TriggerSchema":
        """Creates a TriggerSchema object from a TriggerRequest.

        Args:
            trigger_request: A TriggerRequest object.

        Returns:
            A TriggerSchema object.
        """
        extra_fields = trigger_request.get_extra_fields()

        schema = cls(
            name=trigger_request.name,
            project_id=trigger_request.project,
            user_id=trigger_request.user,
            active=trigger_request.active,
            configuration=trigger_request.get_config(),
            flavor=trigger_request.flavor,
            type=trigger_request.type,
            concurrency=trigger_request.concurrency,
        )

        for field_name, value in extra_fields.items():
            setattr(schema, field_name, value)

        return schema

    def update(self, trigger_update: TriggerUpdate) -> "TriggerSchema":
        """Applies update operation (and validations).

        Args:
            trigger_update: A TriggerUpdate object.

        Returns:
            The updated TriggerSchema.
        """
        if trigger_update.active is not None:
            self.active = trigger_update.active

        if trigger_update.name is not None:
            self.name = trigger_update.name

        if trigger_update.concurrency is not None:
            self.concurrency = trigger_update.concurrency

        self.configuration = trigger_update.get_config()

        for field_name, value in trigger_update.get_extra_fields().items():
            setattr(self, field_name, value)

        return self

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> TRIGGER_RETURN_TYPE_UNION:
        """Converts to Pydantic response model.

        Args:
            include_metadata: Flag - to include metadata.
            include_resources: Flag - include resources.
            **kwargs: Keyword arguments

        Returns:
            A TriggerResponse object.
        """
        body_cls = TYPE_TO_RESPONSE_BODY_MAPPING[self.type]
        response_cls = TYPE_TO_RESPONSE_MAPPING[self.type]

        body = body_cls(
            user_id=self.user_id,
            project_id=self.project_id,
            active=self.active,
            updated=self.updated,
            created=self.created,
            is_archived=self.is_archived,
            type=TriggerType(self.type),
            flavor=TriggerFlavor(self.flavor),
            name=self.name,
            concurrency=self.concurrency,
            **json.loads(self.configuration),
        )

        for field in body.get_extra_fields():
            setattr(body, field, getattr(self, field))

        metadata = None
        if include_metadata:
            metadata = TriggerResponseMetadata()

        resources = None
        if include_resources:
            latest_run = self.latest_run

            resources = TriggerResponseResources(
                user=self.user.to_model() if self.user else None,
                snapshots=[
                    s.to_model(
                        include_resources=False,
                        include_metadata=False,
                        include_config_schema=False,
                        include_python_packages=False,
                    )
                    for s in self.snapshots
                ],
                latest_run=latest_run.to_model()
                if latest_run is not None
                else None,
            )

        return response_cls(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Tekton orchestrator."""

import os
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

import kfp
import requests
import urllib3
from kfp import dsl
from kfp.client import Client as KFPClient
from kfp.compiler import Compiler as KFPCompiler
from kfp_server_api.exceptions import ApiException
from kubernetes import client as k8s_client
from kubernetes import config as k8s_config

from zenml.config.resource_settings import ResourceSettings
from zenml.entrypoints import StepEntrypointConfiguration
from zenml.enums import StackComponentType
from zenml.environment import Environment
from zenml.integrations.tekton.flavors.tekton_orchestrator_flavor import (
    TektonOrchestratorConfig,
    TektonOrchestratorSettings,
)
from zenml.io import fileio
from zenml.logger import get_logger
from zenml.orchestrators import ContainerizedOrchestrator, SubmissionResult
from zenml.orchestrators.utils import get_orchestrator_run_name
from zenml.stack import StackValidator
from zenml.utils import io_utils

if TYPE_CHECKING:
    from zenml.config.base_settings import BaseSettings
    from zenml.models import PipelineRunResponse, PipelineSnapshotResponse
    from zenml.stack import Stack


logger = get_logger(__name__)

ENV_ZENML_TEKTON_RUN_ID = "ZENML_TEKTON_RUN_ID"
KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL = "accelerator"


class KubeClientKFPClient(kfp.Client):  # type: ignore[misc]
    """KFP client initialized from a Kubernetes client.

    This is a workaround for the fact that the native KFP client does not
    support initialization from an existing Kubernetes client.
    """

    def __init__(
        self, client: k8s_client.ApiClient, *args: Any, **kwargs: Any
    ) -> None:
        """Initializes the KFP client from a Kubernetes client.

        Args:
            client: pre-configured Kubernetes client.
            args: standard KFP client positional arguments.
            kwargs: standard KFP client keyword arguments.
        """
        self._k8s_client = client
        super().__init__(*args, **kwargs)

    def _load_config(self, *args: Any, **kwargs: Any) -> Any:
        """Loads the KFP configuration.

        Initializes the KFP configuration from the Kubernetes client.

        Args:
            args: standard KFP client positional arguments.
            kwargs: standard KFP client keyword arguments.

        Returns:
            The KFP configuration.
        """
        from kfp_server_api.configuration import Configuration

        kube_config = self._k8s_client.configuration

        host = (
            kube_config.host
            + "/"
            + self._KUBE_PROXY_PATH.format(kwargs.get("namespace", "kubeflow"))
        )

        config = Configuration(
            host=host,
            api_key=kube_config.api_key,
            api_key_prefix=kube_config.api_key_prefix,
            username=kube_config.username,
            password=kube_config.password,
            discard_unknown_keys=kube_config.discard_unknown_keys,
        )

        # Extra attributes not present in the Configuration constructor
        keys = ["ssl_ca_cert", "cert_file", "key_file", "verify_ssl"]
        for key in keys:
            if key in kube_config.__dict__:
                setattr(config, key, getattr(kube_config, key))

        return config


class TektonOrchestrator(ContainerizedOrchestrator):
    """Orchestrator responsible for running pipelines using Tekton."""

    _k8s_client: Optional[k8s_client.ApiClient] = None

    def _get_kfp_client(
        self,
        settings: TektonOrchestratorSettings,
    ) -> kfp.Client:
        """Creates a KFP client instance.

        Args:
            settings: Settings which can be used to
                configure the client instance.

        Returns:
            A KFP client instance.

        Raises:
            RuntimeError: If the linked Kubernetes connector behaves
                unexpectedly.
        """
        connector = self.get_connector()
        client_args = settings.client_args.copy()

        # The kube_context, host and namespace are stack component
        # configurations that refer to the Tekton deployment. We don't want
        # these overwritten on a run by run basis by user settings
        client_args["namespace"] = self.config.kubernetes_namespace

        if connector:
            client = connector.connect()
            if not isinstance(client, k8s_client.ApiClient):
                raise RuntimeError(
                    f"Expected a k8s_client.ApiClient while trying to use the "
                    f"linked connector, but got {type(client)}."
                )
            return KubeClientKFPClient(
                client=client,
                **client_args,
            )

        elif self.config.kubernetes_context:
            client_args["kube_context"] = self.config.kubernetes_context

        elif self.config.tekton_hostname:
            client_args["host"] = self.config.tekton_hostname

            # Handle username and password, ignore the case if one is passed and
            # not the other. Also do not attempt to get cookie if cookie is
            # already passed in client_args
            if settings.client_username and settings.client_password:
                # If cookie is already set, then ignore
                if "cookie" in client_args:
                    logger.warning(
                        "Cookie already set in `client_args`, ignoring "
                        "`client_username` and `client_password`..."
                    )
                else:
                    session_cookie = self._get_session_cookie(
                        username=settings.client_username,
                        password=settings.client_password,
                    )

                    client_args["cookies"] = session_cookie
        return KFPClient(**client_args)

    def _get_session_cookie(self, username: str, password: str) -> str:
        """Gets session cookie from username and password.

        Args:
            username: Username for tekoton host.
            password: Password for tekoton host.

        Raises:
            RuntimeError: If the cookie fetching failed.

        Returns:
            Cookie with the prefix `authsession=`.
        """
        if self.config.tekton_hostname is None:
            raise RuntimeError(
                "You must configure the tekoton orchestrator "
                "with the `tekton_hostname` parameter which usually ends "
                "with `/pipeline` (e.g. `https://mykubeflow.com/pipeline`). "
                "Please update the current tekoton orchestrator with: "
                f"`zenml orchestrator update {self.name} "
                "--tekton_hostname=<MY_KUBEFLOW_HOST>`"
            )

        # Get cookie
        logger.info(
            f"Attempting to fetch session cookie from {self.config.tekton_hostname} "
            "with supplied username and password..."
        )
        session = requests.Session()
        try:
            response = session.get(self.config.tekton_hostname)
            response.raise_for_status()
        except (
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
            requests.exceptions.RequestException,
        ) as e:
            raise RuntimeError(
                f"Error while trying to fetch tekoton cookie: {e}"
            )

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {"login": username, "password": password}
        try:
            response = session.post(response.url, headers=headers, data=data)
            response.raise_for_status()
        except requests.exceptions.HTTPError as errh:
            raise RuntimeError(
                f"Error while trying to fetch tekoton cookie: {errh}"
            )
        cookie_dict: Dict[str, str] = session.cookies.get_dict()  # type: ignore[no-untyped-call, unused-ignore]

        if "authservice_session" not in cookie_dict:
            raise RuntimeError("Invalid username and/or password!")

        logger.info("Session cookie fetched successfully!")

        return "authservice_session=" + str(cookie_dict["authservice_session"])

    @property
    def config(self) -> TektonOrchestratorConfig:
        """Returns the `TektonOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(TektonOrchestratorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Tekton orchestrator.

        Returns:
            The settings class.
        """
        return TektonOrchestratorSettings

    def get_kubernetes_contexts(self) -> Tuple[List[str], Optional[str]]:
        """Get the list of configured Kubernetes contexts and the active context.

        Returns:
            A tuple containing the list of configured Kubernetes contexts and
            the active context.
        """
        try:
            contexts, active_context = k8s_config.list_kube_config_contexts()
        except k8s_config.config_exception.ConfigException:
            return [], None

        context_names = [c["name"] for c in contexts]
        active_context_name = active_context["name"]
        return context_names, active_context_name

    @property
    def validator(self) -> Optional[StackValidator]:
        """Ensures a stack with only remote components and a container registry.

        Returns:
            A `StackValidator` instance.
        """

        def _validate(stack: "Stack") -> Tuple[bool, str]:
            container_registry = stack.container_registry

            # should not happen, because the stack validation takes care of
            # this, but just in case
            assert container_registry is not None

            kubernetes_context = self.config.kubernetes_context
            msg = f"'{self.name}' Tekton orchestrator error: "

            if not self.connector:
                if not kubernetes_context:
                    return False, (
                        f"{msg}you must either link this stack component to a "
                        "Kubernetes service connector (see the 'zenml "
                        "orchestrator connect' CLI command) or explicitly set "
                        "the `kubernetes_context` attribute to the name of "
                        "the Kubernetes config context pointing to the "
                        "cluster where you would like to run pipelines."
                    )

                contexts, active_context = self.get_kubernetes_contexts()

                if kubernetes_context not in contexts:
                    return False, (
                        f"{msg}could not find a Kubernetes context named "
                        f"'{kubernetes_context}' in the local "
                        "Kubernetes configuration. Please make sure that the "
                        "Kubernetes cluster is running and that the "
                        "kubeconfig file is configured correctly. To list all "
                        "configured contexts, run:\n\n"
                        "  `kubectl config get-contexts`\n"
                    )
                if kubernetes_context != active_context:
                    logger.warning(
                        f"{msg}the Kubernetes context '{kubernetes_context}' "  # nosec
                        f"configured for the Tekton orchestrator is not "
                        f"the same as the active context in the local "
                        f"Kubernetes configuration. If this is not deliberate,"
                        f" you should update the orchestrator's "
                        f"`kubernetes_context` field by running:\n\n"
                        f"  `zenml orchestrator update {self.name} "
                        f"--kubernetes_context={active_context}`\n"
                        f"To list all configured contexts, run:\n\n"
                        f"  `kubectl config get-contexts`\n"
                        f"To set the active context to be the same as the one "
                        f"configured in the Tekton orchestrator and "
                        f"silence this warning, run:\n\n"
                        f"  `kubectl config use-context "
                        f"{kubernetes_context}`\n"
                    )

            silence_local_validations_msg = (
                f"To silence this warning, set the "
                f"`skip_local_validations` attribute to True in the "
                f"orchestrator configuration by running:\n\n"
                f"  'zenml orchestrator update {self.name} "
                f"--skip_local_validations=True'\n"
            )

            if not self.config.is_local:
                # if the orchestrator is not running in a local k3d cluster,
                # we cannot have any other local components in our stack,
                # because we cannot mount the local path into the container.
                # This may result in problems when running the pipeline, "
                # because the local components will not be available inside
                # the Tekton containers.

                # go through all stack components and identify those that
                # advertise a local path where they persist information that
                # they need to be available when running pipelines.
                for stack_comp in stack.components.values():
                    local_path = stack_comp.local_path
                    if not local_path:
                        continue
                    return False, (
                        f"{msg}the Tekton orchestrator is configured to run "
                        f"pipelines in a remote Kubernetes cluster, but the "
                        f"'{stack_comp.name}' {stack_comp.type.value} is a "
                        f"local stack component and will not be available in "
                        f"the Tekton pipeline step.\n"
                        f"Please ensure that you always use non-local "
                        f"stack components with a remote Tekton orchestrator, "
                        f"otherwise you may run into pipeline execution "
                        f"problems. You should use a flavor of "
                        f"{stack_comp.type.value} other than "
                        f"'{stack_comp.flavor}'.\n"
                        + silence_local_validations_msg
                    )

                # if the orchestrator is remote, the container registry must
                # also be remote.
                if container_registry.config.is_local:
                    return False, (
                        f"{msg}the Tekton orchestrator is configured to run "
                        f"pipelines in a remote Kubernetes cluster, but the "
                        f"'{container_registry.name}' container registry URI "
                        f"'{container_registry.config.uri}' "
                        f"points to a local container registry. Please ensure "
                        f"that you always use non-local stack components with "
                        f"a remote Tekton orchestrator, otherwise you will "
                        f"run into problems. You should use a flavor of "
                        f"container registry other than "
                        f"'{container_registry.flavor}'.\n"
                        + silence_local_validations_msg
                    )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate,
        )

    def _create_dynamic_component(
        self,
        image: str,
        command: List[str],
        arguments: List[str],
        component_name: str,
    ) -> dsl.PipelineTask:
        """Creates a dynamic container component for a Tekton pipeline.

        Args:
            image: The image to use for the component.
            command: The command to use for the component.
            arguments: The arguments to use for the component.
            component_name: The name of the component.

        Returns:
            The dynamic container component.
        """

        def dynamic_container_component() -> dsl.ContainerSpec:
            """Dynamic container component.

            Returns:
                The dynamic container component.
            """
            _component = dsl.ContainerSpec(
                image=image,
                command=command,
                args=arguments,
            )

            _component.__name__ = component_name
            return _component

        dynamic_func = FunctionType(
            dynamic_container_component.__code__,
            dynamic_container_component.__globals__,
            name=component_name,
            argdefs=dynamic_container_component.__defaults__,
            closure=dynamic_container_component.__closure__,
        )

        return dsl.container_component(dynamic_func)

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Raises:
            RuntimeError: If you try to run the pipelines in a notebook
                environment.

        Returns:
            Optional submission result.
        """
        # First check whether the code running in a notebook
        if Environment.in_notebook():
            raise RuntimeError(
                "The Tekton orchestrator cannot run pipelines in a notebook "
                "environment. The reason is that it is non-trivial to create "
                "a Docker image of a notebook. Please consider refactoring "
                "your notebook cells into separate scripts in a Python module "
                "and run the code outside of a notebook when using this "
                "orchestrator."
            )

        assert stack.container_registry

        orchestrator_run_name = get_orchestrator_run_name(
            pipeline_name=snapshot.pipeline_configuration.name
        ).replace("_", "-")

        def _create_dynamic_pipeline() -> Any:
            """Create a dynamic pipeline including each step.

            Returns:
                pipeline_func
            """
            step_name_to_dynamic_component: Dict[str, Any] = {}

            for step_name, step in snapshot.step_configurations.items():
                image = self.get_image(
                    snapshot=snapshot,
                    step_name=step_name,
                )
                command = StepEntrypointConfiguration.get_entrypoint_command()
                arguments = (
                    StepEntrypointConfiguration.get_entrypoint_arguments(
                        step_name=step_name,
                        snapshot_id=snapshot.id,
                    )
                )
                dynamic_component = self._create_dynamic_component(
                    image, command, arguments, step_name
                )
                step_settings = cast(
                    TektonOrchestratorSettings, self.get_settings(step)
                )
                node_selector_constraint: Optional[Tuple[str, str]] = None
                pod_settings = step_settings.pod_settings
                if pod_settings:
                    ignored_fields = pod_settings.model_fields_set - {
                        "node_selectors"
                    }
                    if ignored_fields:
                        logger.warning(
                            f"The following pod settings are not supported in "
                            f"Tekton with Tekton Pipelines 2.x and will be "
                            f"ignored: {list(ignored_fields)}."
                        )

                    # apply pod settings
                    if (
                        KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL
                        in pod_settings.node_selectors.keys()
                    ):
                        node_selector_constraint = (
                            KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL,
                            pod_settings.node_selectors[
                                KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL
                            ],
                        )

                step_name_to_dynamic_component[step_name] = dynamic_component

            @dsl.pipeline(  # type: ignore[misc]
                display_name=orchestrator_run_name,
            )
            def dynamic_pipeline() -> None:
                """Dynamic pipeline."""
                # iterate through the components one by one
                # (from step_name_to_dynamic_component)
                for (
                    component_name,
                    component,
                ) in step_name_to_dynamic_component.items():
                    step_environment = step_environments[component_name]
                    # for each component, check to see what other steps are
                    # upstream of it
                    step = snapshot.step_configurations[component_name]
                    upstream_step_components = [
                        step_name_to_dynamic_component[upstream_step_name]
                        for upstream_step_name in step.spec.upstream_steps
                    ]
                    task = (
                        component()
                        .set_display_name(
                            name=component_name,
                        )
                        .set_caching_options(enable_caching=False)
                        .set_env_variable(
                            name=ENV_ZENML_TEKTON_RUN_ID,
                            value=dsl.PIPELINE_JOB_NAME_PLACEHOLDER,
                        )
                        .after(*upstream_step_components)
                    )
                    for key, value in step_environment.items():
                        task = task.set_env_variable(name=key, value=value)
                    self._configure_container_resources(
                        task,
                        step.config.resource_settings,
                        node_selector_constraint,
                    )

            return dynamic_pipeline

        # Get a filepath to use to save the finished yaml to
        fileio.makedirs(self.pipeline_directory)
        pipeline_file_path = os.path.join(
            self.pipeline_directory, f"{orchestrator_run_name}.yaml"
        )

        KFPCompiler().compile(
            pipeline_func=_create_dynamic_pipeline(),
            package_path=pipeline_file_path,
            pipeline_name=orchestrator_run_name,
        )

        logger.info(
            "Writing Tekton workflow definition to `%s`.", pipeline_file_path
        )

        # using the kfp client uploads the pipeline to Tekton pipelines and
        # runs it there
        return self._upload_and_run_pipeline(
            snapshot=snapshot,
            pipeline_file_path=pipeline_file_path,
            run_name=orchestrator_run_name,
        )

    def _upload_and_run_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        pipeline_file_path: str,
        run_name: str,
    ) -> Optional[SubmissionResult]:
        """Tries to upload and run a KFP pipeline.

        Args:
            snapshot: The pipeline snapshot.
            pipeline_file_path: Path to the pipeline definition file.
            run_name: The Tekton run name.

        Raises:
            RuntimeError: If Tekton API returns an error.

        Returns:
            Optional submission result.
        """
        pipeline_name = snapshot.pipeline_configuration.name
        settings = cast(
            TektonOrchestratorSettings, self.get_settings(snapshot)
        )
        user_namespace = settings.user_namespace

        kubernetes_context = self.config.kubernetes_context
        try:
            if kubernetes_context:
                logger.info(
                    "Running in kubernetes context '%s'.",
                    kubernetes_context,
                )
            elif self.config.tekton_hostname:
                logger.info(
                    "Running on Tekton deployment '%s'.",
                    self.config.tekton_hostname,
                )
            elif self.connector:
                logger.info(
                    "Running with Kubernetes credentials from connector '%s'.",
                    str(self.connector),
                )

            # upload the pipeline to Tekton and start it

            client = self._get_kfp_client(settings=settings)
            if snapshot.schedule:
                try:
                    experiment = client.get_experiment(
                        pipeline_name, namespace=user_namespace
                    )
                    logger.info(
                        "A recurring run has already been created with this "
                        "pipeline. Creating new recurring run now.."
                    )
                except (ValueError, ApiException):
                    experiment = client.create_experiment(
                        pipeline_name, namespace=user_namespace
                    )
                    logger.info(
                        "Creating a new recurring run for pipeline '%s'.. ",
                        pipeline_name,
                    )
                logger.info(
                    "You can see all recurring runs under the '%s' experiment.",
                    pipeline_name,
                )

                interval_seconds = (
                    snapshot.schedule.interval_second.seconds
                    if snapshot.schedule.interval_second
                    else None
                )
                result = client.create_recurring_run(
                    experiment_id=experiment.experiment_id,
                    job_name=run_name,
                    pipeline_package_path=pipeline_file_path,
                    enable_caching=False,
                    cron_expression=snapshot.schedule.cron_expression,
                    start_time=snapshot.schedule.utc_start_time,
                    end_time=snapshot.schedule.utc_end_time,
                    interval_second=interval_seconds,
                    no_catchup=not snapshot.schedule.catchup,
                )

                logger.info(
                    "Started recurring run with ID '%s'.",
                    result.recurring_run_id,
                )
            else:
                logger.info(
                    "No schedule detected. Creating a one-off pipeline run.."
                )
                try:
                    result = client.create_run_from_pipeline_package(
                        pipeline_file_path,
                        arguments={},
                        run_name=run_name,
                        enable_caching=False,
                        namespace=user_namespace,
                    )
                except ApiException:
                    raise RuntimeError(
                        f"Failed to create {run_name} on Tekton! "
                        "Please check stack component settings and "
                        "configuration!"
                    )

                logger.info(
                    "Started one-off pipeline run with ID '%s'.", result.run_id
                )

                if settings.synchronous:

                    def _wait_for_completion() -> None:
                        client.wait_for_run_completion(
                            run_id=result.run_id, timeout=settings.timeout
                        )

                    return SubmissionResult(
                        wait_for_completion=_wait_for_completion
                    )
        except urllib3.exceptions.HTTPError as error:
            if kubernetes_context:
                msg = (
                    f"Please make sure your kubernetes config is present and "
                    f"the '{kubernetes_context}' kubernetes context is "
                    "configured correctly."
                )
            elif self.connector:
                msg = (
                    f"Please check that the '{self.connector}' connector "
                    f"linked to this component is configured correctly with "
                    "valid credentials."
                )
            else:
                msg = ""

            logger.warning(
                f"Failed to upload Tekton pipeline: {error}. {msg}",
            )

        return None

    def get_orchestrator_run_id(self) -> str:
        """Returns the active orchestrator run id.

        Raises:
            RuntimeError: If the environment variable specifying the run id
                is not set.

        Returns:
            The orchestrator run id.
        """
        try:
            return os.environ[ENV_ZENML_TEKTON_RUN_ID]
        except KeyError as e:
            raise RuntimeError(
                "Unable to read run id from environment variable "
                f"{ENV_ZENML_TEKTON_RUN_ID}."
            ) from e

    @property
    def root_directory(self) -> str:
        """Returns path to the root directory.

        Returns:
            Path to the root directory.
        """
        return os.path.join(
            io_utils.get_global_config_directory(),
            "tekton",
            str(self.id),
        )

    @property
    def pipeline_directory(self) -> str:
        """Path to a directory in which the Tekton pipeline files are stored.

        Returns:
            Path to the pipeline directory.
        """
        return os.path.join(self.root_directory, "pipelines")

    @property
    def _pid_file_path(self) -> str:
        """Returns path to the daemon PID file.

        Returns:
            Path to the daemon PID file.
        """
        return os.path.join(self.root_directory, "tekton_daemon.pid")

    @property
    def log_file(self) -> str:
        """Path of the daemon log file.

        Returns:
            Path of the daemon log file.
        """
        return os.path.join(self.root_directory, "tekton_daemon.log")

    def _configure_container_resources(
        self,
        dynamic_component: dsl.PipelineTask,
        resource_settings: "ResourceSettings",
        node_selector_constraint: Optional[Tuple[str, str]] = None,
    ) -> dsl.PipelineTask:
        """Adds resource requirements to the container.

        Args:
            dynamic_component: The dynamic component to add the resource
                settings to.
            resource_settings: The resource settings to use for this
                container.
            node_selector_constraint: Node selector constraint to apply to
                the container.

        Returns:
            The dynamic component with the resource settings applied.
        """
        # Set optional CPU, RAM and GPU constraints for the pipeline
        if resource_settings:
            cpu_limit = resource_settings.cpu_count or None

        if cpu_limit is not None:
            dynamic_component = dynamic_component.set_cpu_limit(str(cpu_limit))

        memory_limit = resource_settings.get_memory() or None
        if memory_limit is not None:
            dynamic_component = dynamic_component.set_memory_limit(
                memory_limit
            )

        gpu_limit = (
            resource_settings.gpu_count
            if resource_settings.gpu_count is not None
            else 0
        )

        if node_selector_constraint:
            (constraint_label, value) = node_selector_constraint
            if gpu_limit is not None and gpu_limit > 0:
                dynamic_component = (
                    dynamic_component.set_accelerator_type(value)
                    .set_accelerator_limit(gpu_limit)
                    .set_gpu_limit(gpu_limit)
                )
            elif constraint_label == "accelerator" and gpu_limit == 0:
                logger.warning(
                    "GPU limit is set to 0 but a GPU type is specified. Ignoring GPU settings."
                )

        return dynamic_component



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Vertex utilities."""

import time
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

from google.api_core.exceptions import ServerError
from google.cloud import aiplatform
from google.cloud.aiplatform_v1.types.job_state import JobState

from zenml.enums import ExecutionStatus
from zenml.integrations.gcp.constants import (
    CONNECTION_ERROR_RETRY_LIMIT,
    POLLING_INTERVAL_IN_SECONDS,
    VERTEX_JOB_STATES_COMPLETED,
    VERTEX_JOB_STATES_FAILED,
)
from zenml.integrations.gcp.vertex_custom_job_parameters import (
    VertexCustomJobParameters,
)
from zenml.logger import get_logger

if TYPE_CHECKING:
    from zenml.config.resource_settings import ResourceSettings

logger = get_logger(__name__)


def validate_accelerator_type(accelerator_type: Optional[str] = None) -> None:
    """Validates that the accelerator type is valid.

    Args:
        accelerator_type: The accelerator type to validate.

    Raises:
        ValueError: If the accelerator type is not valid.
    """
    accepted_vals = list(aiplatform.gapic.AcceleratorType.__members__.keys())
    if accelerator_type and accelerator_type.upper() not in accepted_vals:
        raise ValueError(
            f"Accelerator must be one of the following: {accepted_vals}"
        )


def monitor_job(
    job_id: str,
    get_client: Callable[[], aiplatform.gapic.JobServiceClient],
) -> None:
    """Monitors a job until it is completed.

    Args:
        job_id: The ID of the job to monitor.
        get_client: A function that returns an authenticated job service client.

    Raises:
        RuntimeError: If the job fails.
    """
    retry_count = 0
    client = get_client()

    while True:
        time.sleep(POLLING_INTERVAL_IN_SECONDS)
        # Fetch a fresh client in case the credentials have expired
        client = get_client()

        try:
            response = client.get_custom_job(name=job_id)
            retry_count = 0
        except (ConnectionError, ServerError) as err:
            # Retry on connection errors, see also
            # https://github.com/googleapis/google-api-python-client/issues/218
            if retry_count < CONNECTION_ERROR_RETRY_LIMIT:
                retry_count += 1
                logger.warning(
                    f"Error encountered when polling job "
                    f"{job_id}: {err}\nRetrying...",
                )
                continue
            else:
                logger.exception(
                    "Request failed after %s retries.",
                    CONNECTION_ERROR_RETRY_LIMIT,
                )
                raise RuntimeError(
                    f"Request failed after {CONNECTION_ERROR_RETRY_LIMIT} "
                    f"retries: {err}"
                )
        if response.state in VERTEX_JOB_STATES_FAILED:
            err_msg = f"Job `{job_id}` failed: {response}."
            logger.error(err_msg)
            raise RuntimeError(err_msg)
        if response.state in VERTEX_JOB_STATES_COMPLETED:
            break

    logger.info("Job `%s` successful.", job_id)


def build_job_request(
    display_name: str,
    image: str,
    entrypoint_command: List[str],
    custom_job_settings: VertexCustomJobParameters,
    resource_settings: "ResourceSettings",
    environment: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    encryption_spec_key_name: Optional[str] = None,
    service_account: Optional[str] = None,
    network: Optional[str] = None,
    reserved_ip_ranges: Optional[str] = None,
) -> Dict[str, Any]:
    """Build a job request.

    Args:
        display_name: The display name of the job.
        image: The image URI of the job.
        entrypoint_command: The entrypoint command of the job.
        custom_job_settings: The custom job settings.
        resource_settings: The resource settings.
        environment: The environment variables.
        labels: The labels.
        encryption_spec_key_name: The encryption spec key name.
        service_account: The service account.
        network: The network.
        reserved_ip_ranges: The reserved IP ranges.

    Returns:
        Job request dictionary.
    """
    environment = environment or {}
    labels = labels or {}

    validate_accelerator_type(custom_job_settings.accelerator_type)

    accelerator_count = (
        resource_settings.gpu_count or custom_job_settings.accelerator_count
    )
    return {
        "display_name": display_name,
        "job_spec": {
            "worker_pool_specs": [
                {
                    "machine_spec": {
                        "machine_type": custom_job_settings.machine_type,
                        "accelerator_type": custom_job_settings.accelerator_type,
                        "accelerator_count": accelerator_count
                        if custom_job_settings.accelerator_type
                        else 0,
                    },
                    "replica_count": 1,
                    "container_spec": {
                        "image_uri": image,
                        "command": entrypoint_command,
                        "args": [],
                        "env": [
                            {"name": key, "value": value}
                            for key, value in environment.items()
                        ],
                    },
                    "disk_spec": {
                        "boot_disk_type": custom_job_settings.boot_disk_type,
                        "boot_disk_size_gb": custom_job_settings.boot_disk_size_gb,
                    },
                }
            ],
            "service_account": service_account,
            "network": network,
            "reserved_ip_ranges": (
                reserved_ip_ranges.split(",") if reserved_ip_ranges else []
            ),
            "persistent_resource_id": custom_job_settings.persistent_resource_id,
        },
        "labels": labels,
        "encryption_spec": {"kmsKeyName": encryption_spec_key_name}
        if encryption_spec_key_name
        else {},
    }


def convert_job_state(state: JobState) -> ExecutionStatus:
    """Converts a job state to an execution status.

    Args:
        state: The job state.

    Returns:
        The execution status.
    """
    if state in [
        JobState.JOB_STATE_QUEUED,
        JobState.JOB_STATE_PENDING,
    ]:
        return ExecutionStatus.PROVISIONING
    elif state in [
        JobState.JOB_STATE_RUNNING,
        JobState.JOB_STATE_PAUSED,
        JobState.JOB_STATE_UPDATING,
    ]:
        return ExecutionStatus.RUNNING
    elif state == JobState.JOB_STATE_SUCCEEDED:
        return ExecutionStatus.COMPLETED
    elif state == JobState.JOB_STATE_CANCELLING:
        return ExecutionStatus.STOPPING
    elif state == JobState.JOB_STATE_CANCELLED:
        return ExecutionStatus.STOPPED
    elif state in [
        JobState.JOB_STATE_FAILED,
        JobState.JOB_STATE_EXPIRED,
        JobState.JOB_STATE_PARTIALLY_SUCCEEDED,
    ]:
        return ExecutionStatus.FAILED
    else:
        logger.warning(f"Unknown job state: {state}")
        return ExecutionStatus.RUNNING



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""In-memory artifact cache."""

import contextvars
from typing import Any, Dict
from uuid import UUID

from zenml.utils import context_utils


class InMemoryArtifactCache(context_utils.BaseContext):
    """In-memory artifact cache."""

    __context_var__ = contextvars.ContextVar("in_memory_artifact_cache")

    def __init__(self) -> None:
        """Initialize the artifact cache."""
        super().__init__()
        self._cache: Dict[UUID, Any] = {}

    def clear(self) -> None:
        """Clear the artifact cache."""
        self._cache = {}

    def get_artifact_data(self, id_: UUID) -> Any:
        """Get the artifact data.

        Args:
            id_: The ID of the artifact to get the data for.

        Returns:
            The artifact data.
        """
        return self._cache.get(id_)

    def set_artifact_data(self, id_: UUID, data: Any) -> None:
        """Set the artifact data.

        Args:
            id_: The ID of the artifact to set the data for.
            data: The artifact data to set.
        """
        self._cache[id_] = data
